import json
import autograd
import autograd.numpy as np
from functools import wraps, partial
from .util import count_calls, closeleq, closegeq
import scipy
import scipy.optimize
import itertools
import logging

logger = logging.getLogger(__name__)

callback_logger = logging.getLogger(__package__).getChild("progress")


class LoggingCallback(object):

    def __init__(self, user_callback=None):
        self.user_callback = user_callback

    def callback(self, x, fx, i):
        # msg = "Optimization step, {{'it': {i}, 'fun': {fx}, 'x': {x}}}".format(
        #     i=i, x=list(x), fx=fx)
        # logger.info(msg)
        if self.user_callback:
            x = np.array(x)
            x = x.view(LoggingCallbackArray)
            x.iteration = i
            x.fun = fx
            self.user_callback(x)

# special array subclass for LoggingCallback


class LoggingCallbackArray(np.ndarray):
    pass


def _find_minimum(f, start_params, optimizer, bounds=None,
                  callback=None,
                  opt_kwargs={}, **kwargs):
    fixed_params = []
    if bounds is not None:
        bounds = [(None, None) if b is None else b for b in bounds]
        for i, b in enumerate(bounds):
            try:
                if b[0] == b[1] and b[0] is not None:
                    fixed_params += [(i, b[0])]
            except (TypeError, IndexError) as e:
                fixed_params += [(i, b)]

        if any(start_params[i] != b for i, b in fixed_params):
            raise ValueError(
                "start_params does not agree with fixed parameters in bounds")

        if fixed_params:
            fixed_idxs, fixed_offset = list(
                map(np.array, list(zip(*fixed_params))))

            fixed_idxs = np.array([(i in fixed_idxs)
                                   for i in range(len(start_params))])
            proj0 = np.eye(len(fixed_idxs))[:, fixed_idxs]
            proj1 = np.eye(len(fixed_idxs))[:, ~fixed_idxs]

            fixed_offset = np.dot(proj0, fixed_offset)
            get_x = lambda x0: np.dot(proj1, x0) + fixed_offset

            def restricted(fun):
                if fun is None:
                    return None
                new_fun = lambda x0, * \
                    fargs, **fkwargs: fun(get_x(x0), *fargs, **fkwargs)
                return wraps(fun)(new_fun)
            f = restricted(f)
            callback = restricted(callback)

            start_params = np.array(
                [s for (fxd, s) in zip(fixed_idxs, start_params) if not fxd])
            bounds = [b for (fxd, b) in zip(fixed_idxs, bounds) if not fxd]

    opt_kwargs = dict(opt_kwargs)
    assert all([k not in opt_kwargs for k in ['bounds', 'callback']])
    if callback:
        opt_kwargs['callback'] = callback
    if bounds is not None and not all([l is None and h is None for (l, h) in bounds]):
        opt_kwargs['bounds'] = bounds

    ret = _find_minimum_helper(
        f, start_params, optimizer, opt_kwargs, **kwargs)
    if fixed_params:
        ret.x = get_x(ret.x)
    return ret


def _find_minimum_helper(f, start_params, optimizer,
                         opt_kwargs={}, gradmakers={}, replacefun=None):
    opt_kwargs = dict(opt_kwargs)
    for k, v in gradmakers.items():
        assert k not in opt_kwargs
        opt_kwargs[k] = v(f)
    if replacefun:
        f = replacefun(f)
    return optimizer(f, start_params, **opt_kwargs)

stochastic_opts = {}


def is_stoch_opt(fun):
    stochastic_opts[fun.__name__] = fun
    return fun


@is_stoch_opt
def sgd(fun, x0, fun_and_jac, pieces, stepsize, num_iters, bounds=None, callback=None, iter_per_output=10, rgen=np.random):
    x0 = np.array(x0)

    if callback is None:
        callback = lambda *a, **kw: None

    if bounds is None:
        bounds = [(None, None) for _ in x0]
    lower, upper = zip(*bounds)
    lower = [-float('inf') if l is None else l
             for l in lower]
    upper = [float('inf') if u is None else u
             for u in upper]

    def truncate(x):
        return np.maximum(np.minimum(x, upper), lower)

    x = x0
    for nit in range(num_iters):
        i = rgen.randint(pieces)
        f_x, g_x = fun_and_jac(x, i)
        x = truncate(x - stepsize * g_x)
        if nit % iter_per_output == 0:
            callback(x, f_x, nit)

    return scipy.optimize.OptimizeResult({'x': x, 'fun': f_x, 'jac': g_x})


@is_stoch_opt
def adam(fun, x0, fun_and_jac, pieces, num_iters, stepsize=.1, b1=0.9, b2=0.999, eps=10**-8, svrg_epoch=-1, bounds=None, callback=None, rgen=np.random, xtol=1e-6, w=None, fbar=None, gbar=None, checkpoint_file=None, checkpoint_iter=10, start_iter=0, m=None, v=None):
    x0 = np.array(x0)

    if callback is None:
        callback = lambda *a, **kw: None

    if bounds is None:
        bounds = [(None, None) for _ in x0]
    lower, upper = zip(*bounds)
    lower = [-float('inf') if l is None else l
             for l in lower]
    upper = [float('inf') if u is None else u
             for u in upper]

    def truncate(x):
        return np.maximum(np.minimum(x, upper), lower)

    x = x0
    if m is None:
        m = np.zeros(len(x))
    else:
        m = np.array(m)
    if v is None:
        v = np.zeros(len(x))
    else:
        v = np.array(v)

    if w is not None:
        w = np.array(w)
    if gbar is not None:
        gbar = np.array(gbar)

    prev_close = False
    success = False
    for nit in range(start_iter, num_iters):
        i = rgen.randint(pieces)
        f_x, g_x = fun_and_jac(x, i)

        if svrg_epoch > 0 and nit // svrg_epoch and nit % svrg_epoch == 0:
            w = x
            fbar, gbar = fun_and_jac(w, None)
            #logger.info("SVRG pivot, {0}".format(
            #    {"w": list(w), "fbar": fbar, "gbar": list(gbar)}))
        if w is not None:
            f_w, g_w = fun_and_jac(w, i)
            f_x = f_x - f_w + fbar
            g_x = g_x - g_w + gbar

        callback(x, f_x, nit)

        m = (1 - b1) * g_x + b1 * m  # First  moment estimate.
        v = (1 - b2) * (g_x**2) + b2 * v  # Second moment estimate.

        mhat = m / (1 - b1**(nit + 1))    # Bias correction.
        vhat = v / (1 - b2**(nit + 1))

        prev_x = x
        x = truncate(x - stepsize * mhat / (np.sqrt(vhat) + eps))
        #logger.info("Adam moment estimates, {0}".format(
        #    {"x": list(x), "m": list(m), "v": list(v)}))

        # require x to not change for 2 steps in a row before stopping
        if xtol < 0 or not np.allclose(x, prev_x, xtol, xtol):
            prev_close = False
        elif prev_close:
            success = True
            break
        else:
            prev_close = True

        if checkpoint_file is not None and nit % checkpoint_iter == 0:
            def to_list(y):
                if y is None:
                    return None
                else:
                    return list(y)

            with open(checkpoint_file, "w") as f:
                json.dump({
                    "start_iter": nit+1,
                    "fbar": fbar,
                    "gbar": to_list(gbar),
                    "w": to_list(w),
                    "m": to_list(m),
                    "v": to_list(v),
                    "x0": to_list(x)}, f)


    if success:
        message = "|x[k]-x[k-1]|~=0"
    else:
        message = "Maximum number of iterations reached"

    return scipy.optimize.OptimizeResult({'x': x, 'fun': f_x, 'jac': g_x, 'nit': nit, 'message': message,
                                          'success': success})


@is_stoch_opt
def svrg(fun, x0, fun_and_jac, pieces, stepsize, iter_per_epoch, max_epochs=100, bounds=None, callback=None, rgen=np.random, quasinewton=True, init_epoch_svrg=False, xtol=1e-6):
    x0 = np.array(x0)

    if quasinewton is not True and quasinewton is not False:
        init_H = quasinewton
        quasinewton = True
    else:
        init_H = None

    if callback is None:
        callback = lambda *a, **kw: None

    if bounds is None:
        bounds = [(None, None) for _ in x0]
    lower, upper = zip(*bounds)
    lower = [-float('inf') if l is None else l
             for l in lower]
    upper = [float('inf') if u is None else u
             for u in upper]

    def truncate(x):
        return np.maximum(np.minimum(x, upper), lower)

    def update_Hess(H, new_x, prev_x, new_g, prev_g):
        if np.allclose(new_x, prev_x):
            return H

        s = new_x - prev_x
        y = new_g - prev_g
        sy = np.dot(s, y)
        Bs = np.linalg.solve(H, s)

        y_Bs = y - Bs
        if np.abs(np.dot(s, y_Bs)) < 1e-8 * np.linalg.norm(s) * np.linalg.norm(y_Bs):
            # skip SR1 update
            return H

        Hy = np.dot(H, y)
        s_Hy = s - Hy
        H = H + np.outer(s_Hy, s_Hy) / np.dot(s_Hy, y)
        return H

    I = np.eye(len(x0))

    finished = False
    x = x0
    nit = 0
    history = {k: [] for k in ('x', 'f', 'jac')}
    for epoch in itertools.count():
        if epoch > 0:
            prev_w = w
            prev_gbar = gbar
        else:
            prev_w, prev_gbar = None, None

        w = x
        if epoch > 0 or init_epoch_svrg is True:
            fbar, gbar = fun_and_jac(w, None)
            logger.info("SVRG pivot, {0}".format(
                {"w": list(w), "fbar": fbar, "gbar": list(gbar)}))
            #callback(w, fbar, epoch)
            for k, v in (('x', w), ('f', fbar), ('jac', gbar)):
                history[k].append(v)
        elif init_epoch_svrg is False:
            gbar = None
        else:
            fbar, gbar = init_epoch_svrg

        if quasinewton:
            if prev_gbar is not None:
                H = update_Hess(H, w, prev_w, gbar, prev_gbar)
            else:
                assert epoch == 0 or (epoch == 1 and not init_epoch_svrg)
                if init_H is not None:
                    H = init_H
                else:
                    f_w, g_w = fun_and_jac(w, 0)
                    if epoch == 0:
                        u = truncate(w - g_w)
                    else:
                        u = prev_w
                    f_u, g_u = fun_and_jac(u, 0)
                    s, y = (u - w), (g_u - g_w)
                    H = np.abs(np.dot(s, y) / np.dot(y, y)) * I

            H_eigvals, H_eigvecs = scipy.linalg.eigh(H)
            Habs = np.einsum("k,ik,jk->ij",
                             np.abs(H_eigvals),
                             H_eigvecs, H_eigvecs)
            Babs = scipy.linalg.pinvh(Habs)

        if epoch > 0 and xtol >= 0 and np.allclose(w, prev_w, xtol, xtol):
            success = True
            message = "|x[k]-x[k-1]|~=0"
            break
        if epoch >= max_epochs:
            success = False
            message = "Maximum number of iterations reached"
            break

        for k in range(iter_per_epoch):
            i = rgen.randint(pieces)

            f_x, g_x = fun_and_jac(x, i)

            if gbar is not None:
                f_w, g_w = fun_and_jac(w, i)
                g = (g_x - g_w) + gbar
                f = (f_x - f_w) + fbar
            else:
                assert epoch == 0 and init_epoch_svrg is False
                g = g_x
                f = f_x
            callback(x, f, nit)

            if not quasinewton:
                xnext = truncate(x - stepsize * g)
            else:
                xnext = x - stepsize * np.dot(Habs, g)
                if not np.allclose(xnext, truncate(xnext)):
                    model_fun = lambda y: np.dot(
                        g, y - x) + .5 / stepsize * np.dot(y - x, np.dot(Babs, y - x))
                    model_grad = autograd.grad(model_fun)
                    xnext = scipy.optimize.minimize(
                        model_fun, x, jac=model_grad, bounds=bounds).x
                xnext = truncate(xnext)
            x = xnext
            nit += 1

    history = {k: np.array(v) for k, v in history.items()}
    res = scipy.optimize.OptimizeResult({'success': success, 'message': message,
                                         'nit': nit, 'nepoch': epoch, 'history': history,
                                         'x': x, 'fun': fbar, 'jac': gbar})
    if quasinewton:
        res['hess_inv'] = H

    return res



import itertools as it
import autograd.numpy as np
from scipy.special import comb
import scipy.stats
from .compressed_counts import _config2hashable
from ..util import memoize_instance


def build_config_list(sampled_pops, counts, sampled_n=None, ascertainment_pop=None):
    """
    if sampled_n is not None, counts is the derived allele counts

    if sampled_n is None, counts has an extra trailing axis:
      counts[...,0] is ancestral allele count,
      counts[...,1] is derived allele count
    """
    if sampled_n is not None:
        sampled_n = np.array(sampled_n, dtype=int)
        counts1 = np.array(counts, dtype=int, ndmin=2)
        counts0 = sampled_n - counts1
        counts = np.array([counts0, counts1], dtype=int)
        counts = np.transpose(counts, axes=[1, 2, 0])
    counts = np.array(counts, ndmin=3, dtype=int)
    assert counts.shape[1:] == (len(sampled_pops), 2)
    counts.setflags(write=False)
    return ConfigList(sampled_pops, counts, sampled_n, ascertainment_pop)


def build_full_config_list(sampled_pops, sampled_n, ascertainment_pop=None):
    sampled_n = np.array(sampled_n)
    if ascertainment_pop is None:
        ascertainment_pop = [True] * len(sampled_pops)
    ascertainment_pop = np.array(ascertainment_pop)

    ranges = [list(range(n + 1)) for n in sampled_n]
    config_list = []
    for x in it.product(*ranges):
        x = np.array(x, dtype=int)
        if not (np.all(x[ascertainment_pop] == 0) or np.all(
                x[ascertainment_pop] == sampled_n[ascertainment_pop])):
            config_list.append(x)
    return build_config_list(
        sampled_pops, np.array(config_list, dtype=int), sampled_n,
        ascertainment_pop=ascertainment_pop)


class ConfigList(object):
    """
    Stores a list of configs. Important methods/attributes:

    ConfigList.sampled_pops: the population labels
    ConfigList[i] : the i-th config in the list
    ConfigList.sampled_n : the number of alleles sampled per population.
                        used to construct the likelihood vectors for
                        junction tree algorithm.
    """
    def __init__(self, sampled_pops, conf_arr, sampled_n=None,
                 ascertainment_pop=None):
        """Use build_config_list() instead of calling this constructor directly"""
        # If sampled_n=None, ConfigList.sampled_n will be the max number of
        # observed individuals/alleles per population.
        self.sampled_pops = tuple(sampled_pops)
        self.value = conf_arr

        if ascertainment_pop is None:
            ascertainment_pop = [True] * len(sampled_pops)
        self.ascertainment_pop = np.array(ascertainment_pop)
        self.ascertainment_pop.setflags(write=False)
        if all(not a for a in self.ascertainment_pop):
            raise ValueError(
                "At least one of the populations must be used for "
                "ascertainment of polymorphic sites")

        max_n = np.max(np.sum(self.value, axis=2), axis=0)

        if sampled_n is None:
            sampled_n = max_n
        sampled_n = np.array(sampled_n)
        if np.any(sampled_n < max_n):
            raise ValueError("config greater than sampled_n")
        self.sampled_n = sampled_n
        if not np.sum(sampled_n[self.ascertainment_pop]) >= 2:
            raise ValueError("The total sample size of the ascertainment "
                             "populations must be >= 2")

        config_sampled_n = np.sum(self.value, axis=2)
        self.has_missing_data = np.any(config_sampled_n != self.sampled_n)

        if np.any(np.sum(self.value[:, self.ascertainment_pop, :], axis=1)
                  == 0):
            raise ValueError("Monomorphic sites not allowed. In addition, all"
                             " sites must be polymorphic when restricted to"
                             " the ascertainment populations")

    @property
    def config_array(self):
        return self.value

    def __getitem__(self, args):
        return self.value.__getitem__(args)

    def __len__(self): return len(self.value)

    def __eq__(self, other):
        conf_arr = self.value
        try:
            return np.all(conf_arr == other.value)
        except AttributeError:
            return False

    @memoize_instance
    def as_tuple(self):
        ret = []
        for c in self:
            c = tuple(map(tuple, c))
            ret.append(c)
        return tuple(ret)

    def count_subsets(self, derived_weights_dict, total_counts_dict):
        """
        For each config, count number of subsets with given sample
        size, weighted according to the number of derived alleles.

        Parameters
        total_counts_dict: dict mapping pop to n_pop
        derived_weights_dict: dict mapping pop to list of floats
           with length n_pop+1, giving the weight for each
           derived allele count

        Returns
        numpy.ndarray of weighted counts for each config
        """
        assert (set(derived_weights_dict.keys())
                <= set(total_counts_dict.keys()))

        ret = np.ones(len(self))
        for p, n in total_counts_dict.items():
            i = self.sampled_pops.index(p)
            if p in derived_weights_dict:
                curr = np.zeros(len(self))
                assert len(derived_weights_dict[p]) == n+1
                for d, w in enumerate(derived_weights_dict[p]):
                    if w == 0:
                        continue
                    a = n - d
                    curr += w * (comb(self.value[:, i, 0], a)
                                 * comb(self.value[:, i, 1], d))
                ret *= curr
            else:
                ret *= comb(self.value[:, i, :].sum(axis=1), n)
        return ret

    def subsample_probs(self, subconfig):
        """
        Returns the probability of subsampling subconfig
        from each config.
        """
        subconfig = np.array(subconfig)
        total_counts_dict = {p: n for p, n in zip(self.sampled_pops,
                                                  subconfig.sum(axis=1))
                             if n > 0}
        derived_counts_dict = {p: [0]*(n+1)
                               for p, n in total_counts_dict.items()}
        for p, d in zip(self.sampled_pops, subconfig[:, 1]):
            if p in derived_counts_dict:
                derived_counts_dict[p][d] = 1

        num = self.count_subsets(derived_counts_dict, total_counts_dict)
        denom = self.count_subsets({}, total_counts_dict)

        # avoid 0/0
        assert np.all(num[denom == 0] == 0)
        denom[denom == 0] = 1
        return num / denom

    # TODO: remove this method (and self.sampled_n attribute)
    def _copy(self, sampled_n=None):
        """
        Notes
        -----
        Note that momi.expected_sfs, momi.composite_log_likelihood require
        Demography.sampled_n == ConfigList.sampled_n.
        If this is not the case, you can use _copy() to create a copy with the
        correct sampled_n.
        Note this has no affect on the actual allele counts, as missing data
        is allowed. sampled_n is just used to construct (and store) certain
        vectors for the SFS algorithm.
        """
        if sampled_n is None:
            sampled_n = self.sampled_n
        return ConfigList(self.sampled_pops, self.value, sampled_n=sampled_n,
                           ascertainment_pop=self.ascertainment_pop)

    def _vecs_and_idxs(self, folded):
        augmented_configs = self._augmented_configs(folded)
        augmented_idxs = self._augmented_idxs(folded)

        # construct the vecs
        vecs = [np.zeros((len(augmented_configs), n + 1))
                for n in self.sampled_n]

        for i in range(len(vecs)):
            n = self.sampled_n[i]
            derived = np.einsum(
                "i,j->ji", np.ones(len(augmented_configs)), np.arange(n + 1))
            curr = scipy.stats.hypergeom.pmf(
                    k=augmented_configs[:, i, 1],
                    M=n,
                    n=derived,
                    N=augmented_configs[:, i].sum(1)
                )
            assert not np.any(np.isnan(curr))
            vecs[i] = np.transpose(curr)

        # copy augmented_idxs to make it safe
        return vecs, dict(augmented_idxs)

    # def _config_str_iter(self):
    #     for c in self.value:
    #         yield _config2hashable(c)

    def _augmented_configs(self, folded):
        return self._build_augmented_configs_idxs(folded)[0]

    def _augmented_idxs(self, folded):
        return self._build_augmented_configs_idxs(folded)[1]

    @memoize_instance
    def _build_augmented_configs_idxs(self, folded):
        augmented_configs = []
        augmented_config_2_idx = {}  # maps config -> row in vecs

        def augmented_idx(config):
            hashed = _config2hashable(config)
            try:
                return augmented_config_2_idx[hashed]
            except KeyError:
                idx = len(augmented_configs)
                assert idx == len(augmented_config_2_idx)

                augmented_config_2_idx[hashed] = idx
                augmented_configs.append(config)
                return idx

        # initial row is a "zero" config
        null_idx = augmented_idx(
            np.array([(-1, 1)] * len(self.sampled_pops), dtype=int))

        # next row is normalization constant
        denom_idx = augmented_idx(np.zeros((len(self.sampled_pops), 2)))

        # get row indices for each config
        idx_2_row = np.array(list(map(augmented_idx, self)), dtype=int)

        # remove monomorphic configs
        # (if there is missing data or error matrices,
        # expected_sfs_tensor_prod will return nonzero SFS
        # for monomorphic configs)
        monomorphic = np.any(np.sum(self.value, axis=1) == 0, axis=1)
        idx_2_row[monomorphic] = null_idx

        # get row indices for each denominator
        sample_sizes_array = np.sum(self.value, axis=2)
        if np.any(sample_sizes_array > self.sampled_n):
            raise Exception("There is a config that is larger than the"
                            " specified sample size!")

        sample_sizes = [tuple(s) for s in sample_sizes_array]
        # corrections for monomorphic sites (all ancestral & all derived)
        ssize_2_corrections = [{}, {}]
        for s in set(sample_sizes):
            # add rows for monomorphic correction terms
            for mono_allele in (0, 1):
                mono_config = tuple(ss if asc else 0 for ss,
                                    asc in zip(s, self.ascertainment_pop))
                mono_config = np.array(
                    [mono_config, [0] * len(mono_config)], dtype=int, ndmin=2)
                if mono_allele == 1:
                    mono_config = mono_config[::-1, :]
                mono_config = np.transpose(mono_config)
                ssize_2_corrections[mono_allele][
                    s] = augmented_idx(mono_config)
        corrections_2_denom = [np.array([corr_row[s] for s in sample_sizes],
                                        dtype=int)
                               for corr_row in ssize_2_corrections]

        # get row indices for folded configs
        if folded:
            rev_confs = self.value[:, :, ::-1]
            is_symm = np.all(self.value == rev_confs, axis=(1, 2))
            folded_2_row = []
            for rc, symm in zip(rev_confs, is_symm):
                if symm:
                    # map to 0 if symmetric
                    folded_2_row += [null_idx]
                else:
                    folded_2_row += [augmented_idx(rc)]
            folded_2_row = np.array(folded_2_row, dtype=int)
            # dont use monomorphic configs
            folded_2_row[monomorphic] = null_idx

        idxs = {'denom_idx': denom_idx, 'idx_2_row': idx_2_row}
        assert len(corrections_2_denom) == 2
        idxs.update({('corrections_2_denom', 0): corrections_2_denom[0],
                     ('corrections_2_denom', 1): corrections_2_denom[1]})
        try:
            idxs['folded_2_row'] = folded_2_row
        except UnboundLocalError:
            pass

        return np.array(augmented_configs, dtype=int), idxs


class _ConfigList_Subset(ConfigList):
    # Efficient access to subset of configs
    def __init__(self, configs, sub_idxs):
        self.sub_idxs = sub_idxs
        self.full_configs = configs
        for a in ("sampled_n", "sampled_pops",
                  "has_missing_data", "ascertainment_pop"):
            setattr(self, a, getattr(self.full_configs, a))

    @property
    def value(self):
        return self.full_configs.value[self.sub_idxs, :, :]

    def __getitem__(self, args):
        if isinstance(args, tuple):
            arg0 = self.sub_idxs[args[0]]
            rest = tuple([slice(None)] + list(args[1:]))
            return self.full_configs[arg0][rest]
        else:
            return self.full_configs[self.sub_idxs[args]]

    def __iter__(self):
        for i in self.value:
            yield i

    def __len__(self):
        return len(self.sub_idxs)

    # def _vecs_and_idxs(self, folded):
    #     vecs,_ = self.full_configs._vecs_and_idxs(folded)
    #     old_idxs, idxs = self._build_idxs(folded)

    #     vecs = [v[old_idxs,:] for v in vecs]
    #     ## copy idxs to make it safe
    #     return vecs, dict(idxs)

    def _augmented_configs(self, folded):
        return self.full_configs._augmented_configs(
            folded)[self._build_old_new_idxs(folded)[0], :, :]

    def _augmented_idxs(self, folded):
        return self._build_old_new_idxs(folded)[1]

    @memoize_instance
    def _build_old_new_idxs(self, folded):
        idxs = self.full_configs._augmented_idxs(folded)

        denom_idx_key = 'denom_idx'
        denom_idx = idxs[denom_idx_key]
        idxs = {k: v[self.sub_idxs]
                for k, v in list(idxs.items()) if k != denom_idx_key}

        old_idxs = np.array(
            list(set(sum(map(list, idxs.values()), [denom_idx]))))
        old_2_new_idxs = {old_id: new_id for new_id,
                          old_id in enumerate(old_idxs)}

        idxs = {k: np.array([old_2_new_idxs[old_id]
                             for old_id in v], dtype=int)
                for k, v in list(idxs.items())}
        idxs[denom_idx_key] = old_2_new_idxs[denom_idx]
        return old_idxs, idxs



import momi
from momi import expected_sfs, expected_total_branch_len
import random
import itertools
import networkx as nx

from autograd.numpy import log
import autograd.numpy as np

import scipy
import scipy.stats

from demo_utils import simple_admixture_demo, random_tree_demo, simple_admixture_3pop


def check_num_snps(sampled_n_dict, demo, num_loci, mut_rate, ascertainment_pop=None, error_matrices=None):
    if error_matrices is not None or ascertainment_pop is not None:
        # TODO
        raise NotImplementedError

    #seg_sites = momi.simulate_ms(
    #    ms_path, demo, num_loci=num_loci, mut_rate=mut_rate)
    #sfs = seg_sites.sfs

    num_bases = 1000
    sfs = demo.simulate_data(
        sampled_n_dict=sampled_n_dict,
        muts_per_gen=mut_rate/num_bases,
        recoms_per_gen=0,
        length=num_bases,
        num_replicates=num_loci)._sfs

    n_sites = sfs.n_snps(vector=True)

    n_sites_mean = np.mean(n_sites)
    n_sites_sd = np.std(n_sites)

    # TODO this test isn't very useful because expected_branchlen is not used anywhere internally anymore
    n_sites_theoretical = demo.expected_branchlen(sampled_n_dict) * mut_rate
    #n_sites_theoretical = momi.expected_total_branch_len(
    #    demo, ascertainment_pop=ascertainment_pop, error_matrices=error_matrices) * mut_rate

    zscore = -np.abs(n_sites_mean - n_sites_theoretical) / n_sites_sd
    pval = scipy.stats.norm.cdf(zscore) * 2.0

    assert pval >= .05


def test_admixture_3pop_numsnps():
    check_num_snps({"a":1,"b":2,"c":3}, simple_admixture_3pop(),
                   1000, 1.0)
    #check_num_snps({"a":1,"b":2,"c":3}, simple_admixture_3pop(),
    #               1000, 1.0,
    #               ascertainment_pop=[True, True, False])


def test_tree_demo_numsnps():
    lins_per_pop = 10
    num_leaf_pops = 3

    demo = random_tree_demo(num_leaf_pops, lins_per_pop)
    check_num_snps(dict(zip(demo.leafs, [10]*num_leaf_pops)),
                   demo, 1000, 1.0)


def test_admixture_demo_numsnps():
    check_num_snps({"a":1,"b":2}, simple_admixture_demo(), 1000, 1.0)


def check_demo_normalization(demo, ascertainment_pop=None, error_matrices=True, **kwargs):
    #demo = demo.demo_hist._get_multipop_moran(demo.pops, demo.n)
    leaves = demo.sampled_pops

    sampled_n = demo.sampled_n

    config_list = momi.data.configurations.build_full_config_list(
        demo.sampled_pops, sampled_n, ascertainment_pop)
    config_list = config_list._copy(sampled_n=demo.sampled_n)

    sfs = expected_sfs(demo, config_list, normalized=True, **kwargs)
    assert np.isclose(np.sum(sfs), 1.0)

    sfs = expected_sfs(demo, config_list, **kwargs)
    assert np.isclose(np.sum(sfs), expected_total_branch_len(
        demo, ascertainment_pop=ascertainment_pop, **kwargs))

    # check sums to 1 even with error matrix
    error_matrices = [np.exp(np.random.randn(n + 1, n + 1))
                      for n in demo.sampled_n]
    error_matrices = [
        np.einsum('ij,j->ij', x, 1. / np.sum(x, axis=0)) for x in error_matrices]

    sfs = expected_sfs(demo, config_list, normalized=True,
                       error_matrices=error_matrices, **kwargs)
    assert np.isclose(np.sum(sfs), 1.0)

    # check sums to total branch len, even with error matrix
    config_list = config_list._copy(sampled_n=sampled_n)
    error_matrices = None
    if error_matrices:
        error_matrices = [np.exp(np.random.randn(n + 1, n + 1))
                          for n in demo.sampled_n]
        error_matrices = [
            np.einsum('ij,j->ij', x, 1. / np.sum(x, axis=0)) for x in error_matrices]
    else:
        error_matrices = None
    if ascertainment_pop is None:
        ascertainment_pop = [True] * len(demo.sampled_n)
    ascertainment_pop = np.array(ascertainment_pop)

    sfs = expected_sfs(demo, config_list,
                       error_matrices=error_matrices, **kwargs)
    total_len = expected_total_branch_len(
        demo, error_matrices=error_matrices, ascertainment_pop=ascertainment_pop)
    assert np.isclose(np.sum(sfs), total_len)


def test_admixture_3pop_ascertainment():
    check_demo_normalization(simple_admixture_3pop()._get_demo({"a":5,"b":5,"c":5}), ascertainment_pop=[
                             True, True, False])


def test_tree_demo_normalization():
    lins_per_pop = 10
    num_leaf_pops = 3

    demo = random_tree_demo(num_leaf_pops, lins_per_pop)
    check_demo_normalization(demo._get_demo(dict(zip(demo.leafs, [lins_per_pop]*num_leaf_pops))))


def test_admixture_demo_normalization():
    check_demo_normalization(simple_admixture_demo()._get_demo({"a":3,"b":3}), error_matrices=False)



import gzip
import sys
import logging
import argparse
from .data.snps import SnpAlleleCounts


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--verbose", action="store_true")
    parser.add_argument(
        "out", type=str, help="Output file")
    parser.add_argument(
        "n_blocks", type=int, default=1,
        help="Number of blocks for jackknife/bootstrap")
    parser.add_argument(
        "files", nargs="+",
        help="Files containing SNP allele counts")

    args = parser.parse_args()

    if args.verbose:
        logging.basicConfig(level=logging.INFO, format="%(asctime)s %(message)s")

    if len(args.files) == 1:
        f, = args.files
        counts = SnpAlleleCounts.load(f)
    else:
        counts = SnpAlleleCounts.concatenate(
            SnpAlleleCounts.load(fname)
            for fname in args.files)

    logging.info("Extracting SFS...")
    counts.extract_sfs(args.n_blocks).dump(args.out)



import json
import functools
import logging
import time
import autograd.numpy as np
import scipy
import autograd as ag
from autograd.extend import primitive, defvjp
from .optimizers import _find_minimum, stochastic_opts, LoggingCallback
from .compute_sfs import expected_sfs, expected_total_branch_len, expected_heterozygosity
from .demography import Demography
from .data.configurations import _ConfigList_Subset
from .data.sfs import Sfs

logger = logging.getLogger(__name__)


class SfsLikelihoodSurface(object):
    def __init__(self, data, demo_func=None, mut_rate=None, length=1, log_prior=None, folded=False, error_matrices=None, truncate_probs=1e-100, batch_size=1000, p_missing=0.0, use_pairwise_diffs=False):
        """
        Object for computing composite likelihoods, and searching for the maximum composite likelihood.

        Parameters
        ==========
        data: momi.SegSites or momi.Sfs object
        demo_func: function or None
            function that creates momi.Demography object from a vector of input parameters
            if None, this is the identity function (so the "input parameter" is the Demography object itself)
        mut_rate: float or list or None
            The mutation rate.
            If an array, the length should be the total number of loci.
            If None, use a multinomial model instead of Poisson model (so the numger of segregating sites is fixed)
        length: float or array
            The length of each locus.
            The per-locus mutation rate is mut_rate * length.
        folded, error_matrices:
            see help(momi.expected_sfs)
        log_prior:
            function that adds a log-prior to the log-likelihood
        truncate_probs:
            if truncate_probs=eps, then eps is added to the SFS entries before computing the log-likelihood
            this can be helpful for dealing with very small probabilities, within computer accuracy of 0.
        batch_size:
            controls the memory usage. the SFS will be computed in batches of batch_size.
            Decrease batch_size to decrease memory usage (but add running time overhead).
            set batch_size=-1 to compute all SNPs in a single batch. This is required if you wish to compute hessians or higher-order derivatives with autograd.
        processes:
            the number of cores to use.
            if <= 0 (the default), do not use any parallelization.
            if processes > 0, it is recommended to create the SfsLikelihoodSurface()
            using the with...as... construct:

                with SfsLikelihoodSurface(data, demo_func, processes=10) as surface:
                     mle = surface.find_mle(x0)
                print("MLE is ", mle.x)

            as this will automatically take care of closing connections to the parallel subprocesses.
            (Alternatively, you can manually call surface.close(), but care must be taken
            to make sure surface.close() is called in the event of an Error).
        """
        self.data = data

        try:
            self.sfs = self.data.sfs
        except AttributeError:
            self.sfs = self.data

        self.demo_func = demo_func

        self.mut_rate = mut_rate
        if self.mut_rate is not None:
            self.mut_rate = self.mut_rate * np.array(length)
        self.folded = folded
        if self.folded:
            self.sfs = self.sfs.fold()  # required for entropy term to be correct
        self.error_matrices = error_matrices

        self.truncate_probs = truncate_probs

        self.log_prior = log_prior
        self.batch_size = batch_size

        if batch_size <= 0:
            self.sfs_batches = None
        else:
            self.sfs_batches = _build_sfs_batches(self.sfs, batch_size)

        self.p_missing = p_missing

        self.use_pairwise_diffs = use_pairwise_diffs

        if self.mut_rate and self.sfs.configs.has_missing_data and not self.use_pairwise_diffs:
            raise ValueError(
                "Expected total branch length not implemented for missing data; set use_pairwise_diffs=True to scale total mutations by the pairwise differences instead.")

    def log_lik(self, x, vector=False):
        """
        Returns the composite log-likelihood of the data at the point x.
        """
        ret = self._log_lik(x, vector=vector)
        logger.debug("log-likelihood = {0}".format(ret))
        return ret

    def _score(self, x):
        return ag.grad(self.log_lik)(x)

    def _fisher(self, x):
        return -ag.hessian(self.log_lik)(x)

    def _score_cov(self, params):
        params = np.array(params)

        def f_vec(x):
            ret = self._log_lik(x, vector=True)
            # centralize
            return ret - np.mean(ret)

        j = ag.jacobian(f_vec)(params)
        return np.einsum('ij, ik', j, j)

    def _log_lik(self, x, vector):
        demo = self._get_multipop_moran(x)
        ret = self._get_multinom_loglik(demo, vector=vector) + self._mut_factor(demo, vector=vector)
        if vector:
            ret = ret + self._log_prior(x) / len(ret)
        else:
            ret = ret + self._log_prior(x)
        return ret

    def _get_multipop_moran(self, x):
        if self.demo_func:
            logger.debug(
                "Computing log-likelihood at x = {0}".format(str(x).replace('\n', '')))
            demo = self.demo_func(*x)
        else:
            demo = x
        return demo

    def _get_multinom_loglik(self, demo, vector):
        if self.sfs_batches:
            G = demo._get_graph_structure()
            cache = demo._get_differentiable_part()
            ret = 0.0
            for batch in self.sfs_batches:
                ret = ret + _raw_log_lik(
                    cache, G, batch,
                    self.truncate_probs, self.folded,
                    self.error_matrices, vector)
        else:
            ret = _composite_log_likelihood(
                self.data, demo, truncate_probs=self.truncate_probs,
                folded=self.folded, error_matrices=self.error_matrices,
                use_pairwise_diffs=self.use_pairwise_diffs,
                vector=vector)
        return ret

    def _mut_factor(self, demo, vector):
        if self.mut_rate is not None:
            return _mut_factor(
                self.sfs, demo, self.mut_rate,
                vector, self.p_missing, self.use_pairwise_diffs)
        else:
            return 0

    def _log_prior(self, x):
        if self.log_prior:
            return self.log_prior(x)
        else:
            return 0

    def kl_div(self, x):
        """
        Returns KL-Divergence(Empirical || Theoretical(x)).
        """
        log_lik = self.log_lik(x)
        #ret = -log_lik + self.sfs.n_snps() * self.sfs._entropy + _entropy_mut_term(self.mut_rate, self.sfs, self.p_missing, self.use_pairwise_diffs)
        ret = -log_lik + self.sfs.n_snps() * self.sfs._entropy
        if self.mut_rate:
            ret = ret + \
                self.sfs._get_muts_poisson_entropy(self.use_pairwise_diffs)
        ret = ret / float(self.sfs.n_snps())
        if not self.log_prior:
            assert ret >= 0, "kl-div: %s, log_lik: %s, total_count: %s" % (
                str(ret), str(log_lik), str(self.sfs.n_snps()))
        return ret

    def find_mle(self, x0, method="tnc", jac=True, hess=False, hessp=False, bounds=None, callback=None, **kwargs):
        """
        Search for the maximum of the likelihood surface
        (i.e., the minimum of the KL-divergence).

        Parameters
        ==========
        x0 : numpy.ndarray
             initial guess
        method : str
                 Can be any method from scipy.optimize.minimize()
                 (e.g. "tnc","L-BFGS-B",etc.)
        jac : bool
              If True, compute gradient automatically, and pass into the optimization method.
              If False, don't pass in gradient to the optimization method.
        hess, hessp: bool
              Pass hessian/hessian-vector-product into the optimization method.
              Only implemented for some scipy optimizers, and may have high memory cost.
        bounds : list of pairs [(lower0,higher0),...]
              As in scipy.optimize.minimize.
              If None, then do unbounded optimization.
              If one of (lower,higher) is None, then unbounded in that direction.

              If an element of the list is a number (instead of a pair),
              then the optimizer will fix the parameter at that value,
              and optimize over the other parameters.
        callback: callable, optional
              Called after each iteration as callback(x), like in
              scipy.optimize.minimize. However, x is given 2 additional
              attributes, x.iteration and x.fun, that allow the callback
              function to access the current iteration number and the current
              objective function value.
        **kwargs : additional arguments to pass to scipy.optimize.minimize()

        Notes
        =====
        This is just a wrapper around scipy.optimize.minimize, and takes the same arguments, with the following exceptions:
        1) no "fun" param (this is set to be self.kl_div)
        2) jac, hess, hessp are bools. If True, their respective derivatives are defined using autograd and passed into scipy.optimize.minimize; otherwise, "None" is passed in for the derivatives (in which case scipy may use a numerical derivative if needed)
        """
        print_progress = LoggingCallback(user_callback=callback).callback
        hist = lambda: None
        hist.itr = 0
        hist.recent_vals = []
        starttime = time.time()

        def callback(x):
            for y, fx in reversed(hist.recent_vals):
                if np.allclose(y, x):
                    break
            assert np.allclose(y, x)
            try:
                fx = fx.value
            except AttributeError:
                pass
            print_progress(x, fx, hist.itr)
            hist.itr += 1
            hist.recent_vals = [(x, fx)]

        opt_kwargs = dict(kwargs)
        opt_kwargs["method"] = method

        opt_kwargs['jac'] = jac
        if jac:
            replacefun = ag.value_and_grad
        else:
            replacefun = None

        gradmakers = {}
        if hess:
            gradmakers['hess'] = ag.hessian
        if hessp:
            gradmakers['hessp'] = ag.hessian_vector_product

        @functools.wraps(self.kl_div)
        def fun(x):
            ret = self.kl_div(x)
            hist.recent_vals += [(x, ret)]
            return ret

        return _find_minimum(fun, x0, scipy.optimize.minimize,
                             bounds=bounds, callback=callback,
                             opt_kwargs=opt_kwargs, gradmakers=gradmakers, replacefun=replacefun)


    def stochastic_find_mle(
            self, x0, snps_per_minibatch, stepsize, num_iters,
            bounds=None, callback=None,
            checkpoint_file=None, checkpoint_iter=10,
            svrg_epoch=-1, b1=0.9, b2=0.999, eps=10**-8,
            rgen=np.random):
        """
        Search for maximum likelihood using ADAM-style
        stochastic gradient descent.

        Parameters
        ==========
        x0: numpy.array or str
            The starting point for the optimization.
            If a string, a path to the checkpoint file of
            a previous run.
        snps_per_minibatch: int
            The number of SNPs per minibatch
        stepsize: float
            The stepsize for each step
        bounds: None or list of (lower, upper) pairs
            Bounds for the parameter space
        callback: function
            function to call at every step
        checkpoint_file: None or str
            File to save intermediate progress
        checkpoint_iter: int
            Number of iterations between saving intermediate progress
        logging_freq: int
            Frequency to output current iteration to logging
        svrg_epoch: int
            If positive, use SVRG with given epoch length

            SVRG computes the full likelihood every epoch and uses
            this to improve the accuracy of the stochastic gradient
        b1, b2, eps: float
            Parameters for ADAM algorithm. See ADAM paper.
            Recommended to leave these at the default values.
        rgen: random generator
            By default the numpy random generator, which
            can be set with numpy.random.seed().

            Alternatively, use numpy.random.RandomState to create
            a separate random generator and pass it in here.
        """
        kwargs = {}
        kwargs["stepsize"] = stepsize
        kwargs["num_iters"] = num_iters
        kwargs["b1"] = b1
        kwargs["b2"] = b2
        kwargs["eps"] = eps
        kwargs["svrg_epoch"] = svrg_epoch
        kwargs["checkpoint_file"] = checkpoint_file
        kwargs["checkpoint_iter"] = checkpoint_iter
        kwargs["callback"] = callback
        kwargs["bounds"] = bounds

        if isinstance(x0, str):
            with open(x0) as f:
                kwargs.update(json.load(f))
        else:
            kwargs["x0"] = x0

        return self._stochastic_surfaces(
            snps_per_minibatch=snps_per_minibatch,
            rgen=rgen).find_mle(**kwargs)

    def _get_stochastic_pieces(self, pieces, rgen):
        sfs_pieces = _subsfs_list(self.sfs, pieces, rgen)
        if self.mut_rate is None:
            mut_rate = None
        else:
            mut_rate = self.mut_rate * np.ones(self.sfs.n_loci)
            mut_rate = np.sum(mut_rate) / float(pieces)

        return [SfsLikelihoodSurface(sfs, demo_func=self.demo_func, mut_rate=None,
                                     folded=self.folded, error_matrices=self.error_matrices,
                                     truncate_probs=self.truncate_probs, batch_size=self.batch_size)
                for sfs in sfs_pieces]

    def _stochastic_surfaces(self, n_minibatches=None, snps_per_minibatch=None, rgen=np.random):
        """
        Partitions the data into n_minibatches random subsets ("minibatches") of roughly equal size. It returns a StochasticSfsLikelihoodSurface object, which can be used for stochastic gradient descent.

        Useful methods of StochasticSfsLikelihoodSurface are:
        1) StochasticSfsLikelihoodSurface.find_mle(...): search for the MLE using stochastic gradient descent or SVRG
        2) StochasticSfsLikelihoodSurface.get_minibatch(i): the Sfs corresponding to the i-th minibatch
        3) StochasticSfsLikelihoodSurface.n_minibatches: the number of minibatches
        """
        if (n_minibatches is None) == (snps_per_minibatch is None):
            raise ValueError(
                "Exactly one of n_minibatches, snps_per_minibatch should be specified")
        if snps_per_minibatch is not None:
            n_minibatches = int(
                np.ceil(self.sfs.n_snps() / float(snps_per_minibatch)))
        return StochasticSfsLikelihoodSurface(self, n_minibatches, rgen)


class StochasticSfsLikelihoodSurface(object):

    def __init__(self, full_surface, pieces, rgen):
        try:
            assert pieces > 0 and pieces == int(pieces)
        except (TypeError, AssertionError):
            raise ValueError("pieces should be a positive integer")

        self.pieces = full_surface._get_stochastic_pieces(pieces, rgen)
        self.total_snp_counts = full_surface.sfs._total_freqs
        logger.info("Created {n_batches} minibatches, with an average of {n_snps} SNPs and {n_sfs} unique SFS entries per batch".format(n_batches=len(
            self.pieces), n_snps=full_surface.sfs.n_snps() / float(len(self.pieces)), n_sfs=np.mean([len(piece.sfs.configs) for piece in self.pieces])))

        self.rgen = rgen
        self.full_surface = full_surface

    def get_minibatch(self, i): return self.pieces[i].sfs

    @property
    def n_minibatches(self): return len(self.pieces)

    def avg_neg_log_lik(self, x, i):
        if i is None:
            return -self.full_surface.log_lik(x) / self.full_surface.sfs.n_snps()
        demo = self.full_surface._get_multipop_moran(x)
        ret = -self.pieces[i]._get_multinom_loglik(demo, False) * self.n_minibatches
        ret = ret - self.full_surface._mut_factor(demo, False) - self.full_surface._log_prior(x)
        return ret / self.full_surface.sfs.n_snps()

    def find_mle(self, x0, method="adam", bounds=None, rgen=None, callback=None, **kwargs):
        if not rgen:
            rgen = self.rgen
        callback = LoggingCallback(user_callback=callback).callback

        full_surface = self.full_surface

        opt_kwargs = dict(kwargs)
        opt_kwargs.update({'pieces': self.n_minibatches, 'rgen': rgen})

        return _find_minimum(self.avg_neg_log_lik, x0, optimizer=stochastic_opts[method],
                             bounds=bounds, callback=callback, opt_kwargs=opt_kwargs,
                             gradmakers={'fun_and_jac': ag.value_and_grad})


def _composite_log_likelihood(data, demo, mut_rate=None, truncate_probs=0.0, vector=False, p_missing=None, use_pairwise_diffs=False, **kwargs):
    try:
        sfs = data.sfs
    except AttributeError:
        sfs = data

    sfs_probs = np.maximum(expected_sfs(demo, sfs.configs, normalized=True, **kwargs),
                           truncate_probs)
    log_lik = sfs._integrate_sfs(np.log(sfs_probs), vector=vector)

    # add on log likelihood of poisson distribution for total number of SNPs
    if mut_rate is not None:
        log_lik = log_lik + \
            _mut_factor(sfs, demo, mut_rate, vector,
                        p_missing, use_pairwise_diffs)

    if not vector:
        log_lik = np.squeeze(log_lik)
    return log_lik


def _mut_factor(sfs, demo, mut_rate, vector, p_missing, use_pairwise_diffs):
    if use_pairwise_diffs:
        return _mut_factor_het(sfs, demo, mut_rate, vector, p_missing)
    else:
        return _mut_factor_total(sfs, demo, mut_rate, vector)


def _mut_factor_het(sfs, demo, mut_rate, vector, p_missing):
    mut_rate = mut_rate * np.ones(sfs.n_loci)
    E_het = expected_heterozygosity(
        demo,
        restrict_to_pops=np.array(
            sfs.sampled_pops)[sfs.ascertainment_pop])

    p_missing = p_missing * np.ones(len(sfs.ascertainment_pop))
    p_missing = p_missing[sfs.ascertainment_pop]
    lambd = np.einsum("i,j->ij", mut_rate, E_het * (1.0 - p_missing))

    counts = sfs.avg_pairwise_hets[:, sfs.ascertainment_pop]
    ret = -lambd + counts * np.log(lambd) - scipy.special.gammaln(counts + 1)
    ret = ret * sfs.sampled_n[sfs.ascertainment_pop] / float(
        np.sum(sfs.sampled_n[sfs.ascertainment_pop]))
    if not vector:
        ret = np.sum(ret)
    else:
        ret = np.sum(ret, axis=1)
    return ret


def _mut_factor_total(sfs, demo, mut_rate, vector):
    mut_rate = mut_rate * np.ones(sfs.n_loci)

    if sfs.configs.has_missing_data:
        raise ValueError(
            "Expected total branch length not implemented for missing data; set use_pairwise_diffs=True to scale total mutations by the pairwise differences instead.")
    E_total = expected_total_branch_len(
        demo, sampled_pops=sfs.sampled_pops, sampled_n=sfs.sampled_n, ascertainment_pop=sfs.ascertainment_pop)
    lambd = mut_rate * E_total

    counts = sfs.n_snps(vector=True)
    ret = -lambd + counts * np.log(lambd) - scipy.special.gammaln(counts + 1)
    if not vector:
        ret = np.sum(ret)
    return ret

def rearrange_dict_grad(fun):
    """
    Decorator that allows us to save memory on the forward pass,
    by precomputing the gradient
    """
    @primitive
    def wrapped_fun_helper(xdict, dummy):
        ## ag.value_and_grad() to avoid second forward pass
        ## ag.checkpoint() ensures hessian gets properly checkpointed
        val, grad = ag.checkpoint(ag.value_and_grad(fun))(xdict)
        assert len(val.shape) == 0
        dummy.cache = grad
        return val

    def wrapped_fun_helper_grad(ans, xdict, dummy):
        def grad(g):
            #print("foo")
            return {k:g*v for k,v in dummy.cache.items()}
        return grad
    defvjp(wrapped_fun_helper, wrapped_fun_helper_grad, None)

    @functools.wraps(fun)
    def wrapped_fun(xdict):
        return wrapped_fun_helper(ag.dict(xdict), lambda:None)
    return wrapped_fun

def _raw_log_lik(cache, G, data, truncate_probs, folded, error_matrices, vector=False):
    def wrapped_fun(cache):
        demo = Demography(G, cache=cache)
        return _composite_log_likelihood(data, demo, truncate_probs=truncate_probs, folded=folded, error_matrices=error_matrices, vector=vector)
    if vector:
        return ag.checkpoint(wrapped_fun)(cache)
    else:
        ## avoids second forward pass, and has proper
        ## checkpointing for hessian,
        ## but doesn't work for vectorized output
        return rearrange_dict_grad(wrapped_fun)(cache)


#def _build_sfs_batches(sfs, batch_size):
#    counts = sfs._total_freqs
#    sfs_len = len(counts)
#
#    if sfs_len <= batch_size:
#        return [sfs]
#
#    batch_size = int(batch_size)
#    slices = []
#    prev_idx, next_idx = 0, batch_size
#    while prev_idx < sfs_len:
#        slices.append(slice(prev_idx, next_idx))
#        prev_idx = next_idx
#        next_idx += batch_size
#
#    assert len(slices) == int(np.ceil(sfs_len / float(batch_size)))
#
#    idxs = np.arange(sfs_len, dtype=int)
#    idx_list = [idxs[s] for s in slices]
#    return [_sub_sfs(sfs.configs, counts[idx], subidxs=idx) for idx in idx_list]

def _build_sfs_batches(sfs, batch_size):
    sfs_len = len(sfs.configs)
    if sfs_len <= batch_size:
        return [sfs]

    ret = []
    batch_size = int(batch_size)
    for batch_start in range(0, sfs_len, batch_size):
        next_start = min(batch_start + batch_size, sfs_len)
        ret.append(sfs._subset_configs(np.arange(batch_start,
                                                 next_start)))
    return ret


def _subsfs_list(sfs, n_chunks, rnd):
    n_snps = int(sfs.n_snps())
    logger.debug("Splitting {} SNPs into {} minibatches".format(n_snps, n_chunks))

    logger.debug("Building list of length {}".format(n_snps))
    idxs = np.zeros(n_snps, dtype=int)
    total_counts = np.array(sfs._total_freqs, dtype=int)
    curr = 0
    for i, cnt in enumerate(total_counts):
        idxs[curr:(curr+cnt)] = i
        curr += cnt

    logger.debug("Permuting list of {} SNPs".format(n_snps))
    idxs = rnd.permutation(idxs)

    logger.debug("Splitting permuted SNPs into {} minibatches".format(n_chunks))
    ret = []
    for chunk in range(n_chunks):
        chunk_idxs, chunk_cnts = np.unique(idxs[chunk::n_chunks],
                                           return_counts=True)
        sub_configs = _ConfigList_Subset(sfs.configs, chunk_idxs)
        ret.append(Sfs.from_matrix(
            np.array([chunk_cnts]).T, sub_configs,
            folded=sfs.folded, length=None))
    return ret




import pytest
import momi
from momi.demo_model import DemographicModel
from momi.events import get_event_from_old, LeafEvent, SizeEvent, JoinEvent, PulseEvent, GrowthEvent

from demo_utils import simple_admixture_demo, simple_two_pop_demo, piecewise_constant_demo, simple_five_pop_demo, simple_five_pop_demo, exp_growth_model, exp_growth_0_model

import autograd.numpy as np
import sys
import os
import pickle as pickle

# TODO add a test with archaic leafs

MODELS = [{'demo': simple_admixture_demo, 'nlins': (5, 5), 'params': 7},
          {'demo': simple_two_pop_demo, 'nlins': (5, 8), 'params': 4},
          {'demo': piecewise_constant_demo, 'nlins': (10,), 'params': 9},
          {'demo': simple_five_pop_demo, 'nlins': tuple(
              range(1, 6)), 'params': 30},
          {'demo': exp_growth_model, 'nlins': (10,), 'params': 3},
          {'demo': exp_growth_0_model, 'nlins': (10,), 'params': 2},
          ]
MODELS = {m['demo'].__name__: m for m in MODELS}
# for m in MODELS.values():
#    m['demofunc'] = lambda x: m['demo'](x, m['nlins'])

PICKLE = os.path.join(os.path.dirname(
    os.path.realpath(__file__)), "test_sfs.pickle")


def generate_sfs():
    with open(PICKLE, "rb") as sfs_dict_file:
        sfs_dict = pickle.load(sfs_dict_file)
        ret = []
        for k, v in sfs_dict.items():
            m_name, params, sampled_sfs = k

            n_lin = MODELS[m_name]['nlins']
            demo = MODELS[m_name]['demo'](np.array(params))
            yield m_name, v, demo, sampled_sfs


@pytest.mark.parametrize("m_name,v,demo,sampled_sfs", generate_sfs())
def test_generated_cases(m_name, v, demo, sampled_sfs):
    compute_stats(demo, sampled_sfs, *v)


def compute_stats(demo, sampled_sfs, true_sfs=None, true_branch_len=None):
    sampled_sfs = momi.site_freq_spectrum(demo.leafs, to_dict(sampled_sfs))
    demo.set_data(sampled_sfs, length=1)
    demo.set_mut_rate(1)

    exp_branch_len = demo.expected_branchlen()
    exp_sfs = demo.expected_sfs()

    configs = sorted([tuple(map(tuple, c)) for c in sampled_sfs.configs])
    exp_sfs = np.array([exp_sfs[c] for c in configs])

    # use ms units
    exp_branch_len = exp_branch_len / 4.0 / demo.N_e
    exp_sfs = exp_sfs / 4.0 / demo.N_e

    if true_sfs is not None:
        assert np.allclose(true_sfs, exp_sfs, rtol=1e-4)
    if true_branch_len is not None:
        assert np.allclose(true_branch_len, exp_branch_len, rtol=1e-4)

    return exp_sfs, exp_branch_len


def from_dict(sampled_sfs):
    # make it hashable
    return tuple([tuple(locus.items()) for locus in sampled_sfs])


def to_dict(sampled_sfs):
    # make it a dictionary
    return [dict(locus) for locus in sampled_sfs]

if __name__ == "__main__":
    # TODO check this simulation code still works!
    results = {}
    for m_name, m_val in MODELS.items():
        print("# GENERATING %s" % m_name)
        for i in range(10):
            x = np.random.normal(size=m_val['params'])
            demo = m_val['demo'](x, m_val['nlins'])

            demo.demo_hist = demo.demo_hist.rescaled()

            #seg_sites = simulate_ms(
            #    ms_path, demo.demo_hist._get_multipop_moran(demo.pops, demo.n), num_loci=100, mut_rate=1.0)

            # TODO fix this simulation code!!!
            num_bases = 1000
            mu = 1.
            n_loci = 100
            sfs = demo.demo_hist.simulate_data(
                demo.pops, demo.n,
                mutation_rate=mu/num_bases,
                recombination_rate=0,
                length=num_bases,
                num_replicates=n_loci).sfs

            sampled_sfs = from_dict(sfs.to_dict(vector=True))
            results[(m_name, tuple(x), sampled_sfs)
                    ] = compute_stats(demo, sampled_sfs)
    if len(sys.argv) == 2 and sys.argv[1] == "generate":
        pickle.dump(results, open(PICKLE, "wb"))



from momi import expected_sfs, expected_total_branch_len
import momi
import pytest
import random
import autograd.numpy as np
import scipy
import scipy.stats
import itertools
import sys
import shutil
from collections import Counter

from demo_utils import *

import os

#demo_funcs = {f.__name__: f for f in [simple_admixture_demo, simple_two_pop_demo,
#                                      piecewise_constant_demo, exp_growth_model, simple_admixture_3pop]}

n_lineages = {simple_admixture_demo: (2, 3),
              simple_two_pop_demo: (5, 6),
              piecewise_constant_demo: (10,),
              exp_growth_model: (10,),
              simple_admixture_3pop: (4, 4, 4)}


@pytest.mark.parametrize("k,folded",
                         ((f.__name__, bool(b))
                          for f, b in itertools.product(list(n_lineages.keys()),
                                                            [True, False])))
def test_sfs_counts(k, folded):
    """Test to make sure converting momi demography to ms cmd works"""
    #demo = demo_funcs[k]()
    f = globals()[k]
    demo = f()
    n = n_lineages[f]
    check_sfs_counts(demo, demo.leafs, n, folded=folded)


def check_sfs_counts(demo, sampled_pops, sampled_n, theta=2.5, rho=2.5, num_loci=1000, num_bases=1e5, folded=False):
    #seg_sites = demo.simulate_data(
    #    sampled_pops, sampled_n,
    #    mutation_rate=theta / num_bases,
    #    recombination_rate=rho / num_bases,
    #    length=num_bases,
    #    num_replicates=num_loci,
    #)
    #sfs_list = seg_sites.sfs
    demo.set_mut_rate(muts_per_gen=theta / num_bases)
    data = demo.simulate_data(
        length=num_bases, recoms_per_gen=rho / num_bases,
        num_replicates=num_loci,
        sampled_n_dict=dict(zip(sampled_pops, sampled_n)))
    sfs_list = data.extract_sfs(None)

    if folded:
        # pass
        sfs_list = sfs_list.fold()

    #config_list = sorted(set(sum([sfs.keys() for sfs in sfs_list.loci],[])))

    #sfs_vals, branch_len = expected_sfs(demo, sfs_list.configs, folded=folded), expected_total_branch_len(
    #    demo, sampled_pops=sfs_list.sampled_pops, sampled_n=sfs_list.sampled_n)
    demo.set_data(sfs_list)
    sfs_vals = np.array(list(demo.expected_sfs().values()))
    theoretical = sfs_vals / num_loci

    # observed = np.zeros((len(sfs_list.configs), len(sfs_list.loci)))
    # for j in range(sfs_list.n_loci):
    #     for i,config in enumerate(sfs_list.configs):
    #         observed[i,j] = sfs_list.freq(config,locus=j)
    observed = np.array(sfs_list.freqs_matrix.todense())

    labels = list(sfs_list.configs)

    p_val = my_t_test(labels, theoretical, observed)
    print("p-value of smallest p-value under beta(1,num_configs)\n", p_val)
    cutoff = 0.05
    #cutoff = 1.0
    assert p_val > cutoff


def my_t_test(labels, theoretical, observed, min_samples=25):

    assert theoretical.ndim == 1 and observed.ndim == 2
    assert len(theoretical) == observed.shape[
        0] and len(theoretical) == len(labels)

    n_observed = np.sum(observed > 0, axis=1)
    theoretical, observed = theoretical[
        n_observed > min_samples], observed[n_observed > min_samples, :]
    labels = np.array(list(map(str, labels)))[n_observed > min_samples]
    n_observed = n_observed[n_observed > min_samples]

    runs = observed.shape[1]
    observed_mean = np.mean(observed, axis=1)
    bias = observed_mean - theoretical
    variances = np.var(observed, axis=1)

    t_vals = bias / np.sqrt(variances) * np.sqrt(runs)

    # get the p-values
    abs_t_vals = np.abs(t_vals)
    p_vals = 2.0 * scipy.stats.t.sf(abs_t_vals, df=runs - 1)
    print("# labels, p-values, empirical-mean, theoretical-mean, nonzero-counts")
    toPrint = np.array([labels, p_vals, observed_mean,
                        theoretical, n_observed]).transpose()
    toPrint = toPrint[np.array(toPrint[:, 1], dtype='float').argsort()[
        ::-1]]  # reverse-sort by p-vals
    print(toPrint)

    print("Note p-values are for t-distribution, which may not be a good approximation to the true distribution")

    # p-values should be uniformly distributed
    # so then the min p-value should be beta distributed
    return scipy.stats.beta.cdf(np.min(p_vals), 1, len(p_vals))


if __name__ == "__main__":
    #     demo = Demography.from_ms(1.0," ".join(sys.argv[3:]))
    #     check_sfs_counts(demo, mu=float(sys.argv[2]), num_loci=int(sys.argv[1]))
    demo = simple_admixture_3pop()
    demo = demo.demo_hist._get_multipop_moran(demo.pops, demo.n)
    treeseq = demo.simulate_trees(mutation_rate=1)
    seg_sites = demo.simulate_data(mutation_rate=1)



import gzip
import sys
import logging
import argparse
from .data.snps import SnpAlleleCounts


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--verbose", action="store_true")
    parser.add_argument("files", nargs="+",
                        help="Files containing momi json data to concatenate")

    args = parser.parse_args()

    if args.verbose:
        logging.basicConfig(level=logging.INFO, format="%(asctime)s %(message)s")

    SnpAlleleCounts.concatenate(
        SnpAlleleCounts.load(fname)
        for fname in args.files).dump(sys.stdout)



"""
momi (MOran Models for Inference) is a python package for computing the site frequency spectrum,
a summary statistic commonly used in population genetics, and using it to infer demographic history.

Please refer to examples/tutorial.ipynb for usage & introduction.
"""
from .compute_sfs import expected_sfs, expected_total_branch_len, expected_sfs_tensor_prod, expected_tmrca, expected_deme_tmrca
from .likelihood import SfsLikelihoodSurface
from .confidence_region import ConfidenceRegion
from .data.configurations import build_config_list
from .data.sfs import site_freq_spectrum, Sfs
from .data.tensor import sfs_tensor_prod
from .data.snps import SnpAlleleCounts, snp_allele_counts
from .demo_model import DemographicModel
from .demo_plotter import DemographyPlot
from .sfs_stats import SfsModelFitStats, JackknifeGoodnessFitStat
from .data.convert import sfs_from_dadi




from .util import memoize
import autograd.numpy as np
from autograd.numpy import sum, exp, log
from .math_functions import transformed_expi, expm1d
from scipy.special import comb as binom

from .w_matrix import Wmatrix

class SizeHistory(object):

    def __init__(self, tau, scaled_time):
        '''
        tau = length of time interval (in ms units)
        scaled_time = probability of 2 lineages coalescing within interval
                    = \int 2/N(t) dt, where N(t) = pop size in ms units
        '''
        self.tau = tau
        self.scaled_time = scaled_time

    def sfs(self, n):
        if n == 0:
            return np.array([0.])
        Et_jj = self.etjj(n)
        #assert np.all(Et_jj[:-1] - Et_jj[1:] >= 0.0) and np.all(Et_jj >= 0.0) and np.all(Et_jj <= self.tau)

        ret = np.sum(Et_jj[:, None] * Wmatrix(n), axis=0)

        before_tmrca = self.tau - np.sum(ret * np.arange(1, n) / n)
        # ignore branch length above untruncated TMRCA
        if self.tau == float('inf'):
            before_tmrca = 0.0

        ret = np.concatenate((np.array([0.0]), ret, np.array([before_tmrca])))
        return ret

    # def transition_prob(self, v, axis=0):
    #     return moran_model.moran_action(self.scaled_time, v, axis=axis)


class ConstantHistory(SizeHistory):
    '''Constant size population truncated to time tau.'''

    def __init__(self, tau, N):
        if N <= 0.0:
            raise Exception("N must be positive")
        self.N = N
        super(ConstantHistory, self).__init__(tau, 2.0 * tau / N)

    def etjj(self, n):
        j = np.arange(2, n + 1)

        denom = binom(j, 2) / self.N * 2.0
        if self.tau == float('inf'):
            return 1.0 / denom

        scaled_time = denom * self.tau
        #num = -expm1(-scaled_time)
        #ret = num / denom
        ret = expm1d(-scaled_time) * self.tau
        return ret

    def ms_cmd(self, pop_id, start_time, rescale=1.0):
        return "-en %f %d %f" % (start_time / rescale, pop_id, self.N / rescale)


class ExponentialHistory(SizeHistory):

    def __init__(self, tau, growth_rate, N_bottom):
        assert tau != float('inf')
        # if tau == float('inf'):
        #     ## TODO: make tau=inf work with automatic differentiation
        #     raise Exception("Exponential growth in final epoch not implemented. Try setting final growth rate to the constant 0.0, i.e. -eG t 0.0. Setting final growth rate to a variable (-eG t $0) will break, even if the variable equals 0.0.")
        self.N_bottom, self.growth_rate = N_bottom, growth_rate
        self.total_growth = tau * growth_rate
        self.N_top = N_bottom * exp(-self.total_growth)

        super(ExponentialHistory, self).__init__(
            tau, expm1d(self.total_growth) * tau / self.N_bottom * 2.0)

    def etjj(self, n):
        j = np.arange(2, n + 1)
        jChoose2 = binom(j, 2)

        pow0, pow1 = self.N_bottom / jChoose2 / 2.0, self.total_growth
        ret = -transformed_expi(pow0 * self.growth_rate / exp(pow1))
        ret = ret * exp(-expm1d(pow1) * self.tau / pow0 - pow1)
        ret = ret + transformed_expi(pow0 * self.growth_rate)
        ret = ret * pow0

        return ret

    def ms_cmd(self, pop_id, start_time, rescale=1.0):
        return "-en %f %d %f -eg %f %d %f" % (start_time / rescale, pop_id, self.N_bottom / rescale, start_time / rescale, pop_id, self.growth_rate * rescale)


class PiecewiseHistory(SizeHistory):

    def __init__(self, pieces):
        tau = sum([p.tau for p in pieces])
        self.pieces = pieces
        super(PiecewiseHistory, self).__init__(
            tau, sum([pop.scaled_time for pop in self.pieces]))

    def etjj(self, n):
        j = np.arange(2, n + 1)
        jChoose2 = binom(j, 2)

        ret = np.zeros(len(j))
        noCoalProb = np.ones(len(j))
        for pop in self.pieces:
            ret = ret + noCoalProb * pop.etjj(n)
            if pop.scaled_time != float('inf'):
                noCoalProb = noCoalProb * exp(- pop.scaled_time * jChoose2)
            else:
                assert pop is self.pieces[-1]
        return ret

    def ms_cmd(self, pop_id, start_time, rescale=1.0):
        curr = start_time
        ret = []
        for pop in self.pieces:
            ret += [pop.ms_cmd(pop_id, curr, rescale=rescale)]
            curr += pop.tau
        return " ".join(ret)


# given vector [sfs{n,1},...,sfs{n,n}],
# returns (n-1)x(n-1) matrix whose (ij) entry is sfs{i,j}


def sfs_recurrence(sfs, tau):
    n_max = len(sfs)
    ret = np.zeros((n_max + 1, n_max + 1))
    ret[1:, n_max] = sfs

    for n in range(n_max - 1, 0, -1):
        for k in range(1, n + 1):
            ret[(k, n)] = (ret[(k + 1, n + 1)] * (k + 1) / (n + 1) +
                           ret[(k, n + 1)] * (n + 1 - k) / (n + 1))

    # check accuracy
    assert tau == ret[(1, 1)] or abs(log(tau / ret[(1, 1)])) < 1e-14
    return ret[1:, 1:]




from .compute_sfs import expected_sfs
from .likelihood import _composite_log_likelihood
from .util import memoize_instance, make_constant, check_psd
from .math_functions import inv_psd
import scipy
import scipy.stats
import autograd
import autograd.numpy as np

# stuff for confidence intervals


class _ConfidenceRegion(object):
    def __init__(self, point, score, score_cov, fisher, lik_fun, psd_rtol=1e-8):
        self.point = point
        self.score = score
        self.score_cov = check_psd(score_cov, tol=psd_rtol)
        self.fisher = fisher
        self.lik_fun = lik_fun
        self.psd_rtol = psd_rtol

    @memoize_instance
    def godambe(self, inverse=False):
        """
        Returns Godambe Information.
        If the true params are in the interior of the parameter space,
        the composite MLE will be approximately Gaussian with mean 0,
        and covariance given by the Godambe information.
        """
        fisher_inv = inv_psd(self.fisher, tol=self.psd_rtol)
        ret = check_psd(np.dot(fisher_inv, np.dot(
            self.score_cov, fisher_inv)), tol=self.psd_rtol)
        if not inverse:
            ret = inv_psd(ret, tol=self.psd_rtol)
        return ret

    def wald_intervals(self, lower=.025, upper=.975):
        """
        Marginal wald-type confidence intervals.
        """
        conf_lower, conf_upper = scipy.stats.norm.interval(.95,
                                                           loc=self.point,
                                                           scale=np.sqrt(np.diag(self.godambe(inverse=True))))
        return np.array([conf_lower, conf_upper]).T

    def test(self, null_point, sims=int(1e3), test_type="ratio", alt_point=None, null_cone=None, alt_cone=None, p_only=True):
        """
        Returns p-value for a single or several hypothesis tests.
        By default, does a simple hypothesis test with the log-likelihood ratio.

        Note that for tests on the boundary, the MLE for the null and alternative
        models are often the same (up to numerical precision), leading to a p-value of 1

        Parameters
        ----------
        null_point : array or list of arrays
              the MLE of the null model
              if a list of points, will do a hypothesis test for each point
        sims : the number of Gaussian simulations to use for computing null distribution
              ignored if test_type="wald"
        test_type : "ratio" for likelihood ratio, "wald" for Wald test
              only simple hypothesis tests are implemented for "wald"

              For "ratio" test:
              Note that we set the log likelihood ratio to 0 if the two
              likelihoods are within numerical precision (as defined by numpy.isclose)

              For tests on interior of parameter space, it generally shouldn't happen
              to get a log likelihood ratio of 0 (and hence p-value of 1).
              But this can happen on the interior of the parameter space.

        alt_point : the MLE for the alternative models
              if None, use self.point (the point estimate used for this ConfidenceRegion)
              dimensions should be compatible with null_point
        null_cone, alt_cone : the nested Null and Alternative models
              represented as a list, whose length is the number of parameters
              each entry of the list should be in (None,0,1,-1)
                     None: parameter is unconstrained around the "truth"
                     0: parameter is fixed at "truth"
                     1: parameter can be >= "truth"
                     -1: parameter can be <= "truth"

              if null_cone=None, it is set to (0,0,...,0), i.e. totally fixed
              if alt_cone=None, it is set to (None,None,...), i.e. totally unconstrained
        p_only : bool
              if True, only return the p-value (probability of observing a more extreme statistic)
              if False, return 3 values per test:
                   [0] the p-value: (probability of more extreme statistic)
                   [1] probability of equally extreme statistic (up to numerical precision)
                   [2] probability of less extreme statistic

              [1] should generally be 0 in the interior of the parameter space.
              But on the boundary, the log likelihood ratio will frequently be 0,
              leading to a point mass at the boundary of the null distribution.
        """
        in_shape = np.broadcast(np.array(null_point), np.array(alt_point),
                                np.array(null_cone), np.array(alt_cone)).shape

        null_point = np.array(null_point, ndmin=2)

        if null_cone is None:
            null_cone = [0] * null_point.shape[1]
        null_cone = np.array(null_cone, ndmin=2)

        if alt_point is None:
            alt_point = self.point
        alt_point = np.array(alt_point, ndmin=2)

        if alt_cone is None:
            alt_cone = [None] * null_point.shape[1]
        alt_cone = np.array(alt_cone, ndmin=2)

        b = np.broadcast_arrays(null_point, null_cone, alt_point, alt_cone)
        try:
            assert all(bb.shape[1:] == (len(self.point),) for bb in b)
        except AssertionError:
            raise ValueError("points, cones have incompatible shapes")
        b = [list(map(tuple, x)) for x in b]
        null_point, null_cone, alt_point, alt_cone = b

        if test_type == "ratio":
            sims = np.random.multivariate_normal(
                self.score, self.score_cov, size=sims)

            liks = {}
            for p in list(null_point) + list(alt_point):
                if p not in liks:
                    liks[p] = self.lik_fun(np.array(p))

            sim_mls = {}
            for nc, ac in zip(null_cone, alt_cone):
                if (nc, ac) not in sim_mls:
                    nml, nmle = _project_scores(
                        sims, self.fisher, nc, psd_rtol=self.psd_rtol)
                    aml, amle = _project_scores(
                        sims, self.fisher, ac, psd_rtol=self.psd_rtol, init_vals=nmle)
                    sim_mls[(nc, ac)] = (nml, aml)

            ret = []
            for n_p, n_c, a_p, a_c in zip(null_point, null_cone, alt_point, alt_cone):
                lr = _trunc_lik_ratio(liks[n_p], liks[a_p])
                lr_distn = _trunc_lik_ratio(*sim_mls[(n_c, a_c)])
                ret += [list(map(np.mean, [lr > lr_distn,
                                           lr == lr_distn,
                                           lr < lr_distn]))]
            ret = np.array(ret)
        elif test_type == "wald":
            if np.any(np.array(null_cone) != 0) or any(a_c != tuple([None] * len(self.point)) for a_c in alt_cone):
                raise NotImplementedError(
                    "Only simple tests implemented for wald")

            gdmb = self.godambe(inverse=False)

            resids = np.array(alt_point) - np.array(null_point)
            ret = np.einsum("ij,ij->i", resids,
                            np.dot(resids, gdmb))
            ret = 1. - scipy.stats.chi2.cdf(ret, df=len(self.point))
            ret = np.array([ret, [0] * len(ret), 1. - ret]).T
        else:
            raise NotImplementedError("%s tests not implemented" % test_type)

        if p_only:
            ret = ret[:, 0]
        if len(in_shape) == 1:
            ret = np.squeeze(ret)
        return ret


class ConfidenceRegion(_ConfidenceRegion):
    """
    Constructs asymptotic confidence regions and hypothesis tests,
    using the Limit of Experiments theory.
    """

    def __init__(self, point_estimate, demo_func, data, mut_rate=None, length=1, regime="long", psd_rtol=1e-8, **kwargs):
        """
        Parameters
        ----------
        point_estimate : array
                 a statistically consistent estimate for the true parameters.
                 confidence regions and hypothesis tests are computed for a (shrinking)
                 neighborhood around this point.
        demo_func : function that returns a Demography from parameters
        data : SegSites (or Sfs, if regime="many")
        regime : the limiting regime for the asymptotic confidence region
              if "long", number of loci is fixed, and the length of the loci -> infinity.
                 * uses time series information to estimate covariance structure
                 * requires isinstance(data, SegSites)
                 * loci should be independent. they don't have to be identically distributed
              if "many", the number of loci -> infinity
                 * loci should be independent, and roughly identically distributed
        psd_rtol: for checking if certain matrices (e.g. covariance matrices) are positive semidefinite
              if psd_rtol = epsilon, then we will consider a matrix positive semidefinite if its most
              negative eigenvalue has magnitude less than epsilon * most positive eigenvalue.
        **kwargs : additional arguments passed into composite_log_likelihood
        """
        if regime not in ("long", "many"):
            raise ValueError("Unrecognized regime '%s'" % regime)

        try:
            data = data.seg_sites
        except AttributeError:
            data = data

        if mut_rate is not None:
            mut_rate = mut_rate * length

        self.point = np.array(point_estimate)
        self.demo_func = demo_func
        self.data = data
        self.regime = regime
        self.kwargs = dict(kwargs)
        self.psd_rtol = psd_rtol

        self.score = autograd.grad(self.lik_fun)(self.point)
        self.score_cov = _observed_score_covariance(
            self.regime, self.point, self.data,
            self.demo_func, psd_rtol=self.psd_rtol, mut_rate=mut_rate,
            **self.kwargs)
        self.fisher = _observed_fisher_information(
            self.point, self.data, self.demo_func,
            psd_rtol=self.psd_rtol, assert_psd=False,
            mut_rate=mut_rate, **self.kwargs)

    def lik_fun(self, params, vector=False):
        """Returns composite log likelihood from params"""
        return _composite_log_likelihood(self.data, self.demo_func(*params), vector=vector, **self.kwargs)


def _trunc_lik_ratio(null, alt):
    return (1 - np.isclose(alt, null)) * (null - alt)


def _observed_fisher_information(params, data, demo_func, psd_rtol, assert_psd=True, **kwargs):
    params = np.array(params)
    f = lambda x: _composite_log_likelihood(data, demo_func(*x), **kwargs)
    ret = -autograd.hessian(f)(params)
    if assert_psd:
        try:
            ret = check_psd(ret, tol=psd_rtol)
        except AssertionError:
            raise Exception(
                "Observed Fisher Information is not PSD (either due to numerical instability, or because the parameters are not a local maxima in the interior)")
    return ret


def _observed_score_covariance(method, params, seg_sites, demo_func, psd_rtol, **kwargs):
    if method == "long":
        if "mut_rate" in kwargs:
            raise NotImplementedError(
                "'long' godambe method not implemented for Poisson approximation")
        ret = _long_score_cov(params, seg_sites, demo_func, **kwargs)
    elif method == "many":
        ret = _many_score_cov(params, seg_sites, demo_func, **kwargs)
    else:
        raise Exception("Unrecognized method")

    try:
        ret = check_psd(ret, tol=psd_rtol)
    except AssertionError:
        raise Exception("Numerical instability: score covariance is not PSD")
    return ret


def _many_score_cov(params, data, demo_func, **kwargs):
    params = np.array(params)

    def f_vec(x):
        ret = _composite_log_likelihood(
            data, demo_func(*x), vector=True, **kwargs)
        # centralize
        return ret - np.mean(ret)

    # g_out = einsum('ij,ik', jacobian(f_vec)(params), jacobian(f_vec)(params))
    # but computed in a roundabout way because jacobian implementation is slow
    def _g_out_antihess(x):
        l = f_vec(x)
        lc = make_constant(l)
        return np.sum(0.5 * (l**2 - l * lc - lc * l))
    return autograd.hessian(_g_out_antihess)(params)


def _long_score_cov(params, seg_sites, demo_func, **kwargs):
    if "mut_rate" in kwargs:
        raise NotImplementedError(
            "Currently only implemented for multinomial composite likelihood")
    params = np.array(params)

    configs = seg_sites.sfs.configs
    #_,snp_counts = seg_sites.sfs._idxs_counts(None)
    snp_counts = seg_sites.sfs._total_freqs
    weights = snp_counts / float(np.sum(snp_counts))

    def snp_log_probs(x):
        ret = np.log(expected_sfs(
            demo_func(*x), configs, normalized=True, **kwargs))
        return ret - np.sum(weights * ret)  # subtract off mean

    # g_out = sum(autocov(einsum("ij,ik->ikj",jacobian(idx_series), jacobian(idx_series))))
    # computed in roundabout way, in case jacobian is slow for many snps
    # autocovariance is truncated at sqrt(len(idx_series)), to avoid
    # statistical/numerical issues
    def g_out_antihess(y):
        lp = snp_log_probs(y)
        ret = 0.0
        for l in seg_sites._get_likelihood_sequences(lp):
            L = len(l)
            lc = make_constant(l)

            fft = np.fft.fft(l)
            # (assumes l is REAL)
            assert np.all(np.imag(l) == 0.0)
            fft_rev = np.conj(fft) * np.exp(2 * np.pi *
                                            1j * np.arange(L) / float(L))

            curr = 0.5 * (fft * fft_rev - fft *
                          make_constant(fft_rev) - make_constant(fft) * fft_rev)
            curr = np.fft.ifft(curr)[(L - 1)::-1]

            # make real
            assert np.allclose(np.imag(curr / L), 0.0)
            curr = np.real(curr)
            curr = curr[0] + 2.0 * np.sum(curr[1:int(np.sqrt(L))])
            ret = ret + curr
        return ret
    g_out = autograd.hessian(g_out_antihess)(params)
    g_out = 0.5 * (g_out + np.transpose(g_out))
    return g_out


def _project_scores(simulated_scores, fisher_information, polyhedral_cone, psd_rtol, init_vals=None, method="tnc"):
    """
    Under usual theory, the score is asymptotically
    Gaussian, with covariance == Fisher information.

    For Gaussian location model w~N(x,Fisher^{-1}),
    with location parameter x, the likelihood ratio
    of x vs. 0 is

    LR0 = x*z - x*Fisher*x / 2

    where the "score" z=Fisher*w has covariance==Fisher.

    This function takes a bunch of simulated scores,
    and returns LR0 for the corresponding Gaussian
    location MLEs on a polyhedral cone.

    Input:
    simulated_scores: list or array of simulated values.
        For a correctly specified model in the interior
        of parameter space, this would be Gaussian with
        mean 0 and cov == Fisher information.
        For composite or misspecified model, or model
        on boundary of parameter space, this may have
        a different distribution.
    fisher information: 2d array
    polyhedral_cone: list
        length is the number of parameters
        entries are 0,1,-1,None
            None: parameter is unconstrained
            0: parameter == 0
            1: parameter is >= 0
            -1: parameter is <= 0
    """
    if init_vals is None:
        init_vals = np.zeros(simulated_scores.shape)

    fixed_params = [c == 0 for c in polyhedral_cone]
    if any(fixed_params):
        if all(fixed_params):
            return np.zeros(init_vals.shape[0]), init_vals
        fixed_params = np.array(fixed_params)
        proj = np.eye(len(polyhedral_cone))[~fixed_params, :]
        fisher_information = np.dot(proj, np.dot(fisher_information, proj.T))
        simulated_scores = np.einsum("ij,kj->ik", simulated_scores, proj)
        polyhedral_cone = [c for c in polyhedral_cone if c != 0]
        init_vals = np.einsum("ij,kj->ik", init_vals, proj)

        liks, mles = _project_scores(
            simulated_scores, fisher_information, polyhedral_cone, psd_rtol, init_vals, method)
        mles = np.einsum("ik,kj->ij", mles, proj)
        return liks, mles
    else:
        if all(c is None for c in polyhedral_cone):
            # solve analytically
            try:
                fisher_information = check_psd(
                    fisher_information, tol=psd_rtol)
            except AssertionError:
                raise Exception(
                    "Fisher information is not PSD (optimization problem is unbounded and unconstrained)")
            mles = np.linalg.solve(fisher_information, simulated_scores.T).T
            liks = np.einsum("ij,ij->i", mles, simulated_scores)
            liks = liks - .5 * np.einsum("ij,ij->i", mles,
                                         np.dot(mles, fisher_information))
            return liks, mles

        bounds = []
        for c in polyhedral_cone:
            assert c in (None, -1, 1)
            if c == -1:
                bounds += [(None, 0)]
            elif c == 1:
                bounds += [(0, None)]
            else:
                bounds += [(None, None)]

        assert init_vals.shape == simulated_scores.shape

        def obj(x):
            return -np.dot(z, x) + .5 * np.dot(x, np.dot(fisher_information, x))

        def jac(x):
            return -z + np.dot(fisher_information, x)
        sols = []
        for z, i in zip(simulated_scores, init_vals):
            sols += [scipy.optimize.minimize(obj, i,
                                             method=method, jac=jac, bounds=bounds)]
        liks = np.array([-s.fun for s in sols])
        mles = np.array([s.x for s in sols])
        return liks, mles



import networkx as nx
import autograd.numpy as np
from .data.configurations import ConfigList
from .math_functions import (hypergeom_quasi_inverse,
                             binom_coeffs,
                             _apply_error_matrices,
                             convolve_trailing_axes,
                             sum_trailing_antidiagonals)
from .moran_model import moran_transition


def expected_sfs(
        demography, configs, mut_rate=1.0, normalized=False,
        folded=False, error_matrices=None):
    """
    Expected sample frequency spectrum (SFS) entries for the specified
    demography and configs. The expected SFS is the expected number of
    observed mutations for a configuration. If mutation rate=1, it is
    equivalent to the expected branch length subtended by a configuration.

    Parameters
    ----------
    demography : Demography
    configs : ConfigList
        if configs.folded == True, returns the folded SFS entries
    mut_rate : float
         mutation rate per unit time
    normalized : optional, bool
         if True, mut_rate is ignored, and the SFS is divided by the
         expected total branch length.
         The returned values then represent probabilities, that a given
         mutation will segregate according to the specified configurations.

    Returns
    -------
    sfs : 1d numpy.ndarray
         sfs[j] is the SFS entry corresponding to configs[j]


    Other Parameters
    ----------------
    folded: optional, bool
         if True, return the folded SFS value for each entry
         Default is False.
    error_matrices : optional, sequence of 2-dimensional numpy.ndarray
         length-D sequence, where D = number of demes in demography.
         error_matrices[i] describes the sampling error in deme i as:
         error_matrices[i][j,k] = P(observe j mutants in deme i | k mutants in deme i)
         If error_matrices is not None, then the returned value is adjusted
         to account for this sampling error, in particular the effect it
         has on the total number of observed mutations.

    See Also
    --------
    expected_total_branch_len : sum of all expected SFS entries
    expected_sfs_tensor_prod : compute summary statistics of SFS
    """
    sfs, denom = _expected_sfs(demography, configs, folded, error_matrices)
    if normalized:
        sfs = sfs / denom
    else:
        sfs = sfs * mut_rate
    return sfs


def _expected_sfs(demography, configs, folded, error_matrices):
    if np.any(configs.sampled_n != demography.sampled_n) or np.any(configs.sampled_pops != demography.sampled_pops):
        raise ValueError(
            "configs and demography must have same sampled_n, sampled_pops. Use Demography.copy() or ConfigList.copy() to make a copy with different sampled_n.")

    vecs, idxs = configs._vecs_and_idxs(folded)

    if error_matrices is not None:
        vecs = _apply_error_matrices(vecs, error_matrices)

    vals = expected_sfs_tensor_prod(vecs, demography)

    sfs = vals[idxs['idx_2_row']]
    if folded:
        sfs = sfs + vals[idxs['folded_2_row']]

    denom = vals[idxs['denom_idx']]
    for i in (0, 1):
        denom = denom - vals[idxs[("corrections_2_denom", i)]]

    #assert np.all(np.logical_or(vals >= 0.0, np.isclose(vals, 0.0)))

    return sfs, denom


def expected_total_branch_len(demography, error_matrices=None, ascertainment_pop=None,
                              sampled_pops=None, sampled_n=None):
    """
    The expected total branch length of the sample genealogy.
    Equivalently, the expected number of observed mutations when
    mutation rate=1.

    Parameters
    ----------
    demography : Demography

    Returns
    -------
    total : float-like
         the total expected number of SNPs/branch length

    Other Parameters
    ----------------
    error_matrices : optional, sequence of 2-dimensional numpy.ndarray
         length-D sequence, where D = number of demes in demography.
         error_matrices[i] describes the sampling error in deme i as:
         error_matrices[i][j,k] = P(observe j mutants in deme i | k mutants in deme i)
         If error_matrices is not None, then the returned value is adjusted
         to account for this sampling error, in particular the effect it
         has on the total number of observed mutations.

    See Also
    --------
    expected_sfs : individual SFS entries
    expected_tmrca, expected_deme_tmrca : other interesting statistics
    expected_sfs_tensor_prod : compute general class of summary statistics
    """
    if ascertainment_pop is None:
        ascertainment_pop = [True] * len(demography.sampled_n)
    ascertainment_pop = np.array(ascertainment_pop)

    vecs = [[np.ones(n + 1), [1] + [0] * n, [0] * n + [1]]
            if asc else np.ones((3, n + 1), dtype=float)
            for asc, n in zip(ascertainment_pop, demography.sampled_n)]
    if error_matrices is not None:
        vecs = _apply_error_matrices(vecs, error_matrices)

    ret = expected_sfs_tensor_prod(vecs, demography)
    return ret[0] - ret[1] - ret[2]


def expected_tmrca(demography, sampled_pops=None, sampled_n=None):
    """
    The expected time to most recent common ancestor of the sample.

    Parameters
    ----------
    demography : Demography

    Returns
    -------
    tmrca : float-like

    See Also
    --------
    expected_deme_tmrca : tmrca of subsample within a deme
    expected_sfs_tensor_prod : compute general class of summary statistics
    """
    vecs = [np.ones(n + 1) for n in demography.sampled_n]
    n0 = len(vecs[0]) - 1.0
    vecs[0] = np.arange(n0 + 1) / n0
    return np.squeeze(expected_sfs_tensor_prod(vecs, demography))


def expected_heterozygosity(demography, restrict_to_pops=None,
                            error_matrices=None):
    if restrict_to_pops is None:
        restrict_to_pops = demography.sampled_pops
    configs = np.zeros(
        (len(restrict_to_pops), len(demography.sampled_pops), 2),
        dtype=int)
    for i, pop in enumerate(restrict_to_pops):
        if pop in restrict_to_pops:
            configs[i, demography.sampled_pops.index(pop), :] = 1
    configs = ConfigList(demography.sampled_pops, configs,
                         sampled_n=demography.sampled_n)
    return expected_sfs(demography, configs,
                        error_matrices=error_matrices)


def expected_deme_tmrca(demography, deme, sampled_pops=None, sampled_n=None):
    """
    The expected time to most recent common ancestor, of the samples within
    a particular deme.

    Parameters
    ----------
    demography : Demography
    deme : the deme

    Returns
    -------
    tmrca : float

    See Also
    --------
    expected_tmrca : the tmrca of the whole sample
    expected_sfs_tensor_prod : compute general class of summary statistics
    """
    deme = list(demography.sampled_pops).index(deme)
    vecs = [np.ones(n + 1) for n in demography.sampled_n]

    n = len(vecs[deme]) - 1
    vecs[deme] = np.arange(n + 1) / (1.0 * n)
    vecs[deme][-1] = 0.0

    return np.squeeze(expected_sfs_tensor_prod(vecs, demography))


def expected_sfs_tensor_prod(vecs, demography, mut_rate=1.0, sampled_pops=None):
    """
    Viewing the SFS as a D-tensor (where D is the number of demes), this
    returns a 1d array whose j-th entry is a certain summary statistic of the
    expected SFS, given by the following tensor-vector multiplication:

    res[j] = \sum_{(i0,i1,...)} E[sfs[(i0,i1,...)]] * vecs[0][j,i0] * vecs[1][j, i1] * ...

    where E[sfs[(i0,i1,...)]] is the expected SFS entry for config
    (i0,i1,...), as given by expected_sfs

    Parameters
    ----------
    vecs : sequence of 2-dimensional numpy.ndarray
         length-D sequence, where D = number of demes in the demography.
         vecs[k] is 2-dimensional array, with constant number of rows, and
         with n[k]+1 columns, where n[k] is the number of samples in the
         k-th deme. The row vector vecs[k][j,:] is multiplied against
         the expected SFS along the k-th mode, to obtain res[j].
    demo : Demography
    mut_rate : float
         the rate of mutations per unit time

    Returns
    -------
    res : numpy.ndarray (1-dimensional)
        res[j] is the tensor multiplication of the sfs against the vectors
        vecs[0][j,:], vecs[1][j,:], ... along its tensor modes.

    See Also
    --------
    sfs_tensor_prod : compute the same summary statistics for an observed SFS
    expected_sfs : compute individual SFS entries
    expected_total_branch_len, expected_tmrca, expected_deme_tmrca :
         examples of coalescent statistics that use this function
    """
    # NOTE cannot use vecs[i] = ... due to autograd issues
    sampled_n = [np.array(v).shape[-1] - 1 for v in vecs]
    vecs = [np.vstack([np.array([1.0] + [0.0] * n),  # all ancestral state
                       np.array([0.0] * n + [1.0]),  # all derived state
                       v])
            for v, n in zip(vecs, demography.sampled_n)]

    res = _expected_sfs_tensor_prod(vecs, demography, mut_rate=mut_rate)

    # subtract out mass for all ancestral/derived state
    for k in (0, 1):
        res = res - res[k] * np.prod([l[:, -k] for l in vecs], axis=0)
        assert np.isclose(res[k], 0.0)
    # remove monomorphic states
    res = res[2:]

    return res


def _expected_sfs_tensor_prod(vecs, demography, mut_rate=1.0):
    leaf_states = dict(list(zip(demography.sampled_pops, vecs)))

    res = LikelihoodTensorList.compute_sfs(
        leaf_states, demography)

    return res * mut_rate


class LikelihoodTensorList(object):
    @classmethod
    def compute_sfs(cls, leaf_states, demo):
        liklist = cls(leaf_states, demo)
        for event in nx.dfs_postorder_nodes(
                demo._event_tree):
            liklist._process_event(event)
        assert len(liklist.likelihood_list) == 1
        lik, = liklist.likelihood_list
        return lik.sfs

    def __init__(self, leaf_liks_dict, demo):
        self.likelihood_list = [
            LikelihoodTensor(l, 0, [p])
            for p, l in leaf_liks_dict.items()
        ]
        self.demo = demo

    def _get_likelihoods(self, pop):
        for lik in self.likelihood_list:
            if pop in lik.pop_labels:
                return lik

    def _process_event(self, event):
        e_type = self.demo._event_type(event)
        if e_type == 'leaf':
            self._process_leaf_likelihood(event)
        elif e_type == 'merge_subpops':
            self._process_merge_subpops_likelihood(event)
        elif e_type == 'merge_clusters':
            self._process_merge_clusters_likelihood(event)
        elif e_type == 'pulse':
            self._process_pulse_likelihood(event)
        else:
            raise Exception("Unrecognized event type.")

        for newpop in self.demo._parent_pops(event):
            lik = self._get_likelihoods(newpop)
            lik.make_last_axis(newpop)
            n = lik.get_last_axis_n()
            if n > 0:
                lik.add_last_axis_sfs(self.demo._truncated_sfs(newpop))
                if event != self.demo._event_root:
                    lik.matmul_last_axis(
                        np.transpose(moran_transition(
                            self.demo._scaled_time(newpop), n)))

    def _rename_pop(self, oldpop, newpop):
        self._get_likelihoods(oldpop).rename_pop(
            oldpop, newpop)

    def _process_leaf_likelihood(self, event):
        (pop, idx), = self.demo._parent_pops(event)
        if idx == 0:
            self._rename_pop(pop, (pop, idx))
        else:
            # ghost population
            batch_size = self.likelihood_list[0].liks.shape[0]
            self.likelihood_list.append(LikelihoodTensor(
                np.ones((batch_size, 1)), 0,
                [(pop, idx)]
            ))

    def _in_same_lik(self, pop1, pop2):
        return self._get_likelihoods(pop1) is self._get_likelihoods(pop2)

    def _merge_pops(self, newpopname, child_pops, n=None):
       child_liks = [self._get_likelihoods(p)
                     for p in child_pops]

       for lik, pop in zip(child_liks, child_pops):
           lik.make_last_axis(pop)
           lik.mul_trailing_binoms()

       pop1, pop2 = child_pops
       lik1, lik2 = child_liks
       if lik1 is lik2:
           lik1.sum_trailing_antidiagonals()
       else:
           self.likelihood_list.remove(lik2)
           lik1.convolve_trailing_axes(lik2)

       lik1.mul_trailing_binoms(divide=True)
       lik1.rename_pop(pop1, newpopname)

       if n is not None:
            N = lik1.get_last_axis_n()
            if n < N:
                lik1.matmul_last_axis(
                    hypergeom_quasi_inverse(N, n))

    def _process_merge_clusters_likelihood(self, event):
        child_pops = list(self.demo._child_pops(
            event).keys())
        parent_pop, = self.demo._parent_pops(event)

        assert not self._in_same_lik(*child_pops)
        self._merge_pops(parent_pop, child_pops)

    def _process_merge_subpops_likelihood(self, event):
        child_pops = list(self.demo._child_pops(
            event).keys())
        parent_pop, = self.demo._parent_pops(event)
        n = self.demo._n_at_node(parent_pop)

        assert self._in_same_lik(*child_pops)
        self._merge_pops(parent_pop, child_pops, n=n)

    def _process_pulse_likelihood(self, event):
        parent_pops = self.demo._parent_pops(event)
        child_pops_events = self.demo._child_pops(event)
        assert len(child_pops_events) == 2
        child_pops, child_events = list(zip(*list(child_pops_events.items())))

        recipient, non_recipient, donor, non_donor = self.demo._pulse_nodes(event)
        assert set(parent_pops) == set([donor, non_donor])
        assert set(child_pops) == set([recipient, non_recipient])
        if len(set(child_events)) == 2:
            ## more memory-efficient to do split then join
            admixture_probs, admixture_idxs = self.demo._admixture_prob(recipient)
            admixture_probs_dims = [recipient, non_donor, donor]
            assert set(admixture_probs_dims) == set(admixture_idxs)
            admixture_probs = np.transpose(
                admixture_probs,
                [admixture_idxs.index(i)
                 for i in admixture_probs_dims])

            recipient_lik = self._get_likelihoods(recipient)
            donor_lik = self._get_likelihoods(non_recipient)
            assert donor_lik is not recipient_lik

            recipient_lik.make_last_axis(recipient)
            donor_lik.make_last_axis(non_recipient)

            recipient_lik.admix_trailing_pop(admixture_probs, donor)
            self._merge_pops(donor, [donor, non_recipient])
            self._rename_pop(recipient, non_donor)
        else:
            # in this case, (typically) more memory-efficient to multiply likelihood by transition 4-tensor
            # (if only 2 populations, and much fewer SFS entries than samples, it may be more efficient to replace -ep with -es,-ej)
            child_event, = set(child_events)
            lik = self._get_likelihoods(recipient)
            assert lik is self._get_likelihoods(non_recipient)
            pulse_probs, pulse_idxs = self.demo._pulse_prob(event)
            pulse_probs_dims = [recipient, non_recipient, non_donor, donor]
            assert set(pulse_probs_dims) == set(pulse_idxs)
            pulse_probs = np.transpose(pulse_probs, [
                pulse_idxs.index(i)
                for i in pulse_probs_dims])

            lik.make_last_axis(recipient)
            lik.make_last_axis(non_recipient)

            lik.matmul_last_axis(pulse_probs, axes=2)

            lik.rename_pop(recipient, non_donor)
            lik.rename_pop(non_recipient, donor)


class LikelihoodTensor(object):
    def __init__(self, liks, sfs, pop_labels):
        self.liks = liks
        self.sfs = sfs
        self.pop_labels = pop_labels
        # extra leading dimension for batch
        assert len(self.pop_labels) + 1 == len(
            self.liks.shape)

    def copy(self):
        return LikelihoodTensor(self.liks, self.sfs, list(self.pop_labels))

    def admix_trailing_pop(self, admix_probs_3tensor,
                           newpop_name):
        """
        admix_probs_3tensor should be array returned by demography._admixture_prob_helper
        """
        self.matmul_last_axis(admix_probs_3tensor)
        self.pop_labels.append(newpop_name)

    def sum_trailing_antidiagonals(self):
        trailing_shape = list(self.liks.shape[-2:])
        lik = np.reshape(self.liks, [-1] + trailing_shape)
        lik = sum_trailing_antidiagonals(lik)
        self.liks = np.reshape(lik, [-1] + list(self.liks.shape[1:-2]) + [sum(trailing_shape) - 1])
        self.pop_labels.pop()

    def convolve_trailing_axes(self, other):
        def within_pop_sfs(a, b):
            return a.sfs * b.liks[
                tuple([slice(None)] + [0] * len(b.pop_labels))]
        self.sfs = within_pop_sfs(
            self, other) + within_pop_sfs(other, self)

        class reshape_to_3tensor(object):
            def __init__(self, lik):
                self.lik = np.reshape(lik, [
                    lik.shape[0], -1, lik.shape[-1]])
                self.old_shape = lik.shape

        reshaped0 = reshape_to_3tensor(other.liks)
        reshaped1 = reshape_to_3tensor(self.liks)

        assert reshaped0.lik.shape[0] == reshaped1.lik.shape[0]
        convolved = convolve_trailing_axes(reshaped0.lik, reshaped1.lik)
        self.liks = np.reshape(
            convolved,
            [reshaped0.lik.shape[0]] +
            list(reshaped0.old_shape[1:-1]) +
            list(reshaped1.old_shape[1:-1]) +
            [-1])
        self.pop_labels = list(
            other.pop_labels[:-1]) + list(self.pop_labels)

    @property
    def n_pops(self):
        return len(self.pop_labels)

    @property
    def n_axes(self):
        # extra dimension for the batches (data)
        return self.n_pops + 1

    def pop_axis(self, pop):
        # extra leading dimension for batch (data)
        return self.pop_labels.index(pop) + 1

    def rename_pop(self, oldpop, newpop):
        self.pop_labels[
            self.pop_labels.index(oldpop)] = newpop

    def make_last_axis(self, pop):
        axis = self.pop_axis(pop)
        perm = [i for i in range(self.n_axes)
                if i != axis] + [axis]
        self.liks = np.transpose(self.liks, perm)
        self.pop_labels = [p for p in self.pop_labels if p != pop] + [pop]
        assert len(self.pop_labels) + 1 == len(self.liks.shape)

    def add_last_axis_sfs(self, truncated_sfs):
        self.sfs = self.sfs + np.dot(
            self.liks[tuple(
                [slice(None)] +
                [0] * (self.n_pops - 1) +
                [slice(None)])],
            truncated_sfs)

    def get_last_axis_n(self):
        return self.liks.shape[-1] - 1

    def get_last_axis_binom_coeffs(self):
        return binom_coeffs(self.get_last_axis_n())

    def mul_trailing_binoms(self, divide=False):
        coeffs = self.get_last_axis_binom_coeffs()
        if divide:
            coeffs = 1.0 / coeffs
        self.mul_trailing(coeffs)

    def matmul_last_axis(self, mat, axes=1):
        reshaped_liks = np.reshape(self.liks, [-1] + [np.prod(
            self.liks.shape[-axes:])])
        reshaped_mat = np.reshape(mat, [np.prod(
            mat.shape[:axes], dtype=int)] + [-1])
        reshaped_liks = np.dot(reshaped_liks, reshaped_mat)
        self.liks = np.reshape(reshaped_liks, list(self.liks.shape[:-axes]) + list(mat.shape[axes:]))

    def mul_trailing(self, to_mult):
        self.liks = self.liks * to_mult




import pytest
import autograd.numpy as np
from autograd import grad
import networkx as nx
import random
from demo_utils import *
import momi
from momi import expected_sfs, expected_total_branch_len
from numdifftools import Gradient, Hessian


def check_gradient(f, x):
    print(x, "\n", f(x))

    print("# grad2")
    grad2 = Gradient(f)(x)
    print("# building grad1")
    g = grad(f)
    print("# computing grad1")
    grad1 = g(x)

    print("gradient1\n", grad1, "\ngradient2\n", grad2)
    np.allclose(grad1, grad2)

    # check Hessian vector product
    y = np.random.normal(size=x.shape)
    gdot = lambda u: np.dot(g(u), y)
    hess1, hess2 = grad(gdot)(x), Gradient(gdot)(x)
    print("hess1\n", hess1, "\nhess2\n", hess2)
    np.allclose(hess1, hess2)


def test_simple():
    def f(x):
        return np.sum(np.outer(x, x))
    check_gradient(f, np.random.normal(size=10))


def test_simple_2():
    def f(x):
        return np.sum(1.0 / np.outer(x, x))
    check_gradient(f, np.random.normal(size=10))


def test_simple_3():
    def f(x):
        return np.sum(x) / np.sum(x)
    check_gradient(f, np.random.normal(size=10))


def sfs_func(demo_func, n_lins, normalized=True, states=None):
    # get random sfs entry
    if states is None:
        n = np.sum([x for x in n_lins])
        total_der = 0
        while total_der == 0 or total_der == n:
            total_der = 0
            #states = {}
            states = []
            for pop, n_pop in enumerate(n_lins):
                n_der = random.randint(0, n_pop)
                assert n_der >= 0 and n_der <= n_pop
                total_der += n_der
                #states[pop] = {'ancestral' : n_pop - n_der, 'derived' : n_der}
                states.append(n_der)

    def f(x):
        demo = demo_func(x)
        demo.set_mut_rate(1.0)
        configs = momi.build_config_list(demo.leafs, tuple(states), n_lins)
        #print(configs)
        # print demo.graph['cmd']
        #sfs, branch_len = expected_sfs(demo.demo_hist, configs), expected_total_branch_len(
        #    demo.demo_hist, sampled_pops=demo.pops, sampled_n=demo.n)
        sfs, branch_len = demo.expected_sfs(configs, length=1, return_dict=False), demo.expected_branchlen(
            dict(zip(configs.sampled_pops, configs.sampled_n)))
        if normalized:
            return np.squeeze(sfs / branch_len)
        return np.squeeze(sfs)
    return f


def test_admixture():
    n_lins = (2, 2)
    #f = sfs_func(simple_admixture_demo, n_lins, normalized=True)
    f = sfs_func(simple_admixture_demo, n_lins, normalized=False)
    x = np.random.normal(size=7)
    check_gradient(f, x)


@pytest.mark.parametrize("log_tau,growth_rate,end_growth_rate",
                         ((-random.expovariate(1), g, random.gauss(0, 1))
                          for g in (random.uniform(-10, 10),
                                    random.uniform(-.01, .01), 0.0)))
def test_exp_growth(log_tau, growth_rate, end_growth_rate):
    n_lins = (10,)

    f = sfs_func(exp_growth_model, n_lins, normalized=True)
    x = np.array([log_tau, growth_rate, end_growth_rate])

    check_gradient(f, x)
    #assert False


@pytest.mark.parametrize("n,epochs,normalized",
                         ((random.randint(2, 10), random.randint(1, 10), norm) for norm in (False, True)))
def test_piecewise_constant(n, epochs, normalized):
    n_lins = (n,)

    x = np.random.normal(size=2 * epochs - 1)
    f = sfs_func(piecewise_constant_demo, n_lins, normalized=normalized)

    check_gradient(f, x)


@pytest.mark.parametrize("n1,n2,normalized",
                         ((random.randint(1, 10), random.randint(1, 10), norm) for norm in (False, True)))
def test_simple_two_pop(n1, n2, normalized):
    n_lins = (n1, n2)

    x = np.random.normal(size=4)
    f = sfs_func(simple_two_pop_demo, n_lins, normalized=normalized)

    check_gradient(f, x)


## TODO reenable the test, it got really slow at some point...
#@pytest.mark.parametrize("n1,n2,n3,n4,n5,normalized",
#                         ((random.randint(1, 5), random.randint(1, 5), random.randint(1, 5), random.randint(1, 5), random.randint(1, 5), norm) for norm in (False, True)))
#def test_simple_five_pop(n1, n2, n3, n4, n5, normalized):
#    n_lins = (n1, n2, n3, n4, n5)
#
#    x = np.random.normal(size=30)
#
#    def demo_func(y, n_lins):
#        return simple_five_pop_demo(y, n_lins)
#    f = sfs_func(demo_func, n_lins, normalized=normalized,
#                 states=[0,0,1,0,0])
#
#    check_gradient(f, x)



import autograd.numpy as np
import pandas as pd
from ..math_functions import symmetric_matrix, log_wishart_pdf, slogdet_pos
from ..compute_sfs import expected_sfs_tensor_prod


def sfs_tensor_prod(sfs, vecs):
    """
    Viewing the SFS as a D-tensor (where D is the number of demes), this
    returns a 1d array whose j-th entry is a summary statistic given by the
    tensor-vector multiplication:

    res[j] = \sum_{(i0,i1,...)} sfs[(i0,i1,...)] * vecs[0][j,i0] * vecs[1][j, i1] * ...

    Parameters
    ----------
    sfs : 
    vecs : list of 2-dimensional numpy.ndarray
         a list of length D, where D is the number of demes in the sample.
         vecs[k] is 2-dimensional array, with constant number of rows, and
         with n[k]+1 columns, where n[k] is the number of samples in the
         k-th deme. The row vectors vecs[k][j,:] are multiplied against
         the SFS along the k-th mode, to obtain res[j].

    Returns
    -------
    res : numpy.ndarray (1-dimensional)
        res[j] is the tensor multiplication of the sfs against the vectors
        vecs[0][j,:], vecs[1][j,:], ... along its tensor modes.

    See Also
    --------
    expected_sfs_tensor_prod : compute the expectation of sfs_tensor_prod for a
         randomly sampled sfs.
    """
    res = 0.
    #sfs = sfs.get_dict()
    #entries, counts = list(zip(*list(sfs.items())))
    #entries = np.array(entries, ndmin=3)
    entries = sfs.configs.value
    counts = sfs._total_freqs

    sampled_n = np.array([v.shape[1] - 1 for v in vecs], dtype=int)
    if np.any(np.sum(entries, axis=2) != sampled_n):
        raise NotImplementedError(
            "SFS-tensor-product not implemented for missing data. Consider removing all entries with deficient sample size.")

    for config, val in zip(entries, counts):
        for d, (_, i) in zip(vecs, config):
            val = val * d[:, i]
        res = res + val
    return res

# TODO: rewrite commented code
# try:
#     import sktensor as skt
#     from sktensor.tucker import hosvd
# except ImportError:
#     pass

# def get_sfs_tensor(sfs, n_per_pop):
#     idx, vals = zip(*(sfs.iteritems()))
#     idx = tuple(np.array(x) for x in zip(*idx))
# return skt.sptensor(idx, vals, shape=tuple(n+1 for n in n_per_pop),
# dtype=np.float)

# def greedy_hosvd(sfs_tensor, n_entries, verbose=False):
#     U_list = hosvd(sfs_tensor, sfs_tensor.shape, compute_core=False)
#     total_energy = sfs_tensor.norm()**2
#     curr_entries = [(total_energy, [], [], sfs_tensor)]

#     for d in range(len(sfs_tensor.shape)):
#         prev_entries = curr_entries
#         curr_entries = []

#         for prev_energy, prev_vecs, prev_idxs, prev_tens in prev_entries:

#             energy_sum = 0.0
#             for next_idx, next_dir in enumerate(U_list[d].T):
#                 next_tens = prev_tens.ttv((next_dir,) , (0,) )
#                 try:
#                     energy = next_tens.norm()**2
#                 except AttributeError:
#                     energy = next_tens**2
#                 energy_sum += energy
#                 curr_entries.append( (energy,
#                                       prev_vecs + [next_dir],
#                                       prev_idxs + [next_idx],
#                                       next_tens))

#             curr_entries = sorted(curr_entries, key=lambda x: x[0], reverse=True)[:n_entries]
#             assert np.isclose(energy_sum, prev_energy)
#     if verbose:
#         #print "# Selected components:\n", [idx for _,_,idx,_ in curr_entries]
#         to_print = pd.DataFrame([(idx, energy / total_energy) for energy,_,idx,_ in curr_entries],
#                                 columns=['Component','Percent Energy'])
#         print "# Selected components:\n", to_print, "\n# Unselected percent energy:", 1.0-sum(to_print['Percent Energy'])
#     return [vecs for _,vecs,_,_ in curr_entries]


# class MEstimatorSurface(object):
#     def __init__(self, mu, demo_func=lambda demo: demo):
#         '''
#         demo_func: a function that returns demography from the parameters.
#                    Default is the identity function (so surface is parametrized
#                    by space of demographies).
#                    Note derivatives and covariance estimation only work if
#                    parameters are array-valued.
#         mu: One of the following:
#                (a) a number giving the mutation rate at each locus
#                (b) a list of numbers giving the mutation rate at each locus
#                (c) a function that takes in parameters, and returns (a) or (b)
#                (d) None. In this case uses multinomial probability, instead of Poisson Random Field
#         '''
#         self.mu = mu
#         self.demo_func = demo_func

#     def evaluate(self, params, vector=False):
#         pass

#     def _get_mu(self, params):
#         ## TODO: require same mu for all loci, until I figure out how to appropriately compute confidence intervals when using different mu
#         try:
#             ret = self.mu(params)
#         except TypeError:
#             ret = self.mu
#         return np.array(ret)

# class PoisGaussSurface(MEstimatorSurface):
#     def __init__(self, sfs_list, sfs_directions, mu, demo_func=lambda demo: demo):
#         super(PoisGaussSurface, self).__init__(mu, demo_func)

#         self.n_vecs = sfs_directions[0].shape[0]
#         self.sfs_directions = [np.vstack([[1.0]*s.shape[1], s])
#                                for s in sfs_directions]

#         self.n_loci = len(sfs_list)

#         self.sfs_aggregated = sum_sfs_list(sfs_list)
#         projection = sfs_tensor_prod(self.sfs_aggregated, self.sfs_directions)

#         self.num_muts, projection = projection[0], projection[1:]
#         self.means = projection / self.num_muts

#     def inv_cov_mat(self, demo, branch_len, expectations):
#         pass

#     def evaluate(self, params, vector=False):
#         if vector:
#             raise Exception("Vectorized likelihood not implemented")

#         demo = self.demo_func(params)

#         expectations = expected_sfs_tensor_prod(self.sfs_directions, demo)
#         branch_len, expectations = expectations[0], expectations[1:]
#         expectations = expectations / branch_len

#         mu = self._get_mu(params)
#         mu = np.ones(self.n_loci) * mu
#         mu = np.sum(mu)

#         resids =  expectations - self.means
#         Sigma_inv = self.inv_cov_mat(demo, branch_len, expectations)

# return self.neg_log_lik(mu * branch_len, self.num_muts, resids,
# Sigma_inv)

#     def neg_log_lik(self, expected_snps, n_snps, resids, Sigma_inv):
# return expected_snps - n_snps * np.log(expected_snps) + 0.5 * n_snps *
# np.dot(resids, np.dot(Sigma_inv, resids)) - 0.5 * slogdet_pos(Sigma_inv
# * n_snps)

# def get_cross_vecs(sfs_directions, n_vecs):
#     cross_vecs = []
#     for vecs in sfs_directions:
#         assert n_vecs == vecs.shape[0]

#         idx0, idx1 = np.triu_indices(n_vecs)
#         cross_vecs += [np.einsum('ik,jk->ijk',
#                                  vecs, vecs)[idx0,idx1,:]]
#     return cross_vecs

# class PGSurface_Empirical(PoisGaussSurface):
#     def __init__(self, sfs_list, sfs_directions, mu, demo_func=lambda demo: demo):
#         super(PGSurface_Empirical, self).__init__(sfs_list, sfs_directions, mu, demo_func)

#         cross_means = sfs_tensor_prod(self.sfs_aggregated, get_cross_vecs(sfs_directions, self.n_vecs)) / self.num_muts
#         cross_means = symmetric_matrix(cross_means, self.n_vecs)

#         cov_mat = cross_means - np.outer(self.means, self.means)

#         self.Sigma_inv = np.linalg.inv(cov_mat)

#     def inv_cov_mat(self, demo, branch_len, means):
#         return self.Sigma_inv

# class PGSurface_Diag(PoisGaussSurface):
#     def __init__(self, sfs_list, sfs_directions, mu, demo_func=lambda demo: demo):
#         super(PGSurface_Diag, self).__init__(sfs_list, sfs_directions, mu, demo_func)

#         self.square_sfs_vecs = [s**2 for s in sfs_directions]

#     def inv_cov_mat(self, demo, branch_len, means):
# return np.diag(1./ (expected_sfs_tensor_prod(self.square_sfs_vecs, demo)
# / branch_len - means**2))

# class PGSurface_Exact(PoisGaussSurface):
#     def __init__(self, sfs_list, sfs_directions, mu, demo_func=lambda demo: demo):
#         super(PGSurface_Exact, self).__init__(sfs_list, sfs_directions, mu, demo_func)

#         self.cross_vecs = get_cross_vecs(sfs_directions, self.n_vecs)
#         cross_means = sfs_tensor_prod(self.sfs_aggregated, self.cross_vecs) / self.num_muts
#         cross_means = symmetric_matrix(cross_means, self.n_vecs)

#         self.empirical_covariance = cross_means - np.outer(self.means, self.means)

#     def inv_cov_mat(self, demo, branch_len, means):
#         cross_means = expected_sfs_tensor_prod(self.cross_vecs, demo) / branch_len
#         cross_means = symmetric_matrix(cross_means, self.n_vecs)

#         return np.linalg.inv(cross_means - np.outer(means, means))

# class PoissonWishartSurface(PGSurface_Exact):
#     def neg_log_lik(self, expected_snps, n_snps, resids, Sigma_inv):
#         ret = super(PoissonWishartSurface, self).neg_log_lik(expected_snps, n_snps, resids, Sigma_inv)
#         return ret - log_wishart_pdf(self.empirical_covariance * self.num_muts,
# np.linalg.inv(Sigma_inv), self.num_muts-1, self.n_vecs)




import autograd.numpy as np
#from autograd.core import primitive
from autograd.extend import primitive, defvjp
import scipy
from .util import memoize, check_psd
from .convolution import convolve_sum_axes, transposed_convolve_sum_axes, sum_trailing_antidiagonals, add_trailing_axis, roll_trailing_axes, unroll_trailing_axes
from .einsum2 import einsum1, einsum2


def par_einsum(*args):
    return einsum2(*args)

def convolve_trailing_axes(A, B):
    A = np.reshape(A, list(A.shape) + [1])
    B = np.reshape(B, list(B.shape) + [1])
    return convolve_sum_axes(A, B)

convolve_sum_axes = primitive(convolve_sum_axes)
transposed_convolve_sum_axes = primitive(transposed_convolve_sum_axes)

defvjp(
    convolve_sum_axes,
    lambda ans, A, B: lambda g: transposed_convolve_sum_axes(g, B),
    lambda ans, A, B: lambda g: transposed_convolve_sum_axes(
        np.transpose(g, (0, 2, 1, 3)), A))
#convolve_sum_axes.defgrad(lambda ans, A, B: lambda g: transposed_convolve_sum_axes(g, B))
#convolve_sum_axes.defgrad(lambda ans, A, B: lambda g: transposed_convolve_sum_axes(np.transpose(g, (0, 2, 1, 3)), A), argnum=1)
#convolve_sum_axes.defvjp(
#    lambda g, ans, vs, gvs, A, B: transposed_convolve_sum_axes(g, B))
#convolve_sum_axes.defvjp(
#    lambda g, ans, vs, gvs, A, B: transposed_convolve_sum_axes(
#        np.transpose(g, (0, 2, 1, 3)), A),
#    argnum=1)

defvjp(
    transposed_convolve_sum_axes,
    lambda ans, C, B: lambda g: convolve_sum_axes(g, B),
    lambda ans, C, B: lambda g: transposed_convolve_sum_axes(
        np.transpose(C, (0, 2, 1, 3)), g))
#transposed_convolve_sum_axes.defgrad(lambda ans, C, B: lambda g: convolve_sum_axes(g, B))
#transposed_convolve_sum_axes.defgrad(lambda ans, C, B: lambda g: transposed_convolve_sum_axes(np.transpose(C, (0, 2, 1, 3)), g), argnum=1)
#transposed_convolve_sum_axes.defvjp(
#    lambda g, ans, vs, gvs, C, B: convolve_sum_axes(g, B))
#transposed_convolve_sum_axes.defvjp(
#    lambda g, ans, vs, gvs, C, B: transposed_convolve_sum_axes(
#        np.transpose(C, (0, 2, 1, 3)), g),
#    argnum=1)


def convolve_axes(arr0, arr1, labs, axes, out_axis):
    old_labs = [list(l) for l in labs]
    labs = [[l_i for l_i in l if l_i != a] + [a] for l, a in zip(labs, axes)]

    arr0, arr1 = [einsum1(a, ol, l)
                  for a, ol, l in zip((arr0, arr1), old_labs, labs)]
    reshaped_arrs = [np.reshape(
        a, (a.shape[0], -1, a.shape[-1]), order='C') for a in (arr0, arr1)]
    ret = convolve_trailing_axes(*reshaped_arrs)
    return np.reshape(ret, tuple([ret.shape[0]] + list(arr0.shape[1:-1]) + list(arr1.shape[1:-1]) + [-1]),
                      order='C'), [labs[0][0]] + labs[0][1:-1] + labs[1][1:-1] + [out_axis]

sum_trailing_antidiagonals = primitive(sum_trailing_antidiagonals)
add_trailing_axis = primitive(add_trailing_axis)

defvjp(sum_trailing_antidiagonals,
       lambda ans, A: lambda g: add_trailing_axis(g, A.shape[2]))
#sum_trailing_antidiagonals.defgrad(lambda ans, A: lambda g: add_trailing_axis(g, A.shape[2]))
#sum_trailing_antidiagonals.defvjp(lambda g, ans, vs, gvs, A: add_trailing_axis(g, A.shape[2]))

defvjp(add_trailing_axis, lambda ans, A, trailing_dim: lambda g: sum_trailing_antidiagonals(g))
#add_trailing_axis.defgrad(lambda ans, A, trailing_dim: lambda g: sum_trailing_antidiagonals(g))
#add_trailing_axis.defvjp(lambda g, ans, vs, gvs, A, trailing_dim: sum_trailing_antidiagonals(g))


def sum_antidiagonals(arr, labels, axis0, axis1, out_axis):
    old_labels = list(labels)
    labels = [l for l in labels if l not in (axis0, axis1)]
    arr = einsum1(arr, old_labels, labels + [axis0, axis1])

    reshaped_arr = np.reshape(
        arr, (-1, arr.shape[-2], arr.shape[-1]), order='C')
    ret = sum_trailing_antidiagonals(reshaped_arr)
    return np.reshape(ret, tuple(list(arr.shape[:-2]) + [-1]), order='C'), labels + [out_axis]

roll_trailing_axes = primitive(roll_trailing_axes)
unroll_trailing_axes = primitive(unroll_trailing_axes)

#roll_trailing_axes.defgrad(lambda ans, A: lambda g: unroll_trailing_axes(g))
#roll_trailing_axes.defvjp(lambda g, ans, vs, gvs, A: unroll_trailing_axes(g))
defvjp(roll_trailing_axes,
       lambda ans, A: lambda g: unroll_trailing_axes(g))

#unroll_trailing_axes.defgrad(lambda ans, A: lambda g: roll_trailing_axes(g))
#unroll_trailing_axes.defvjp(lambda g, ans, vs, gvs, A: roll_trailing_axes(g))
defvjp(unroll_trailing_axes,
       lambda ans, A: lambda g: roll_trailing_axes(g))


def roll_axes(arr, labels, axis0, axis1):
    tmp_labels = [l for l in labels if l not in (axis0, axis1)]
    tmp_labels += [axis0, axis1]

    arr = einsum1(arr, labels, tmp_labels)
    reshaped_arr = np.reshape(
        arr, (-1, arr.shape[-2], arr.shape[-1]), order='C')
    ret = roll_trailing_axes(reshaped_arr)
    ret = np.reshape(ret, tuple(list(arr.shape[:-1]) + [-1]), order='C')
    return einsum1(ret, tmp_labels, labels)

'''
Returns
-expi(-1/x) * exp(1/x) / x
for x s.t. abs(x) is decreasing
'''


def transformed_expi(x):
    abs_x = np.abs(x)
    ser = abs_x < 1. / 45.
    nser = np.logical_not(ser)

#     ret = np.zeros(x.shape)
#     ret[ser], ret[nser] = transformed_expi_series(x[ser]), transformed_expi_naive(x[nser])))
#     return ret

    # We use np.concatenate to combine.
    # would be better to use ret[ser] and ret[nser] as commented out above
    # but array assignment not yet supported by autograd
    assert np.all(abs_x[:-1] >= abs_x[1:])
    return np.concatenate((transformed_expi_naive(x[nser]), transformed_expi_series(x[ser])))


def transformed_expi_series(x):
    c_n, ret = 1., 1.
    for n in range(1, 11):
        c_n = -c_n * x * n
        ret = ret + c_n
    return ret


def transformed_expi_naive(x):
    return -expi(-1.0 / x) * np.exp(1.0 / x) / x


@primitive
def expi(x):
    return scipy.special.expi(x)
#expi.defgrad(lambda ans, x: lambda g: g * np.exp(x) / x)
#expi.defvjp(lambda g, ans, vs, gvs, x: g * np.exp(x) / x)
defvjp(expi, lambda ans, x: lambda g: g * np.exp(x) / x)

'''
returns (e^x-1)/x, for scalar x. works for x=0.
Taylor series is 1 + x/2! + x^2/3! + ...
'''


def expm1d(x, eps=1e-6):
    x = np.array(x)
    abs_x = np.abs(x)
    if x.shape:
        # FIXME: don't require abs_x to be increasing
        assert np.all(abs_x[1:] >= abs_x[:-1])
        small = abs_x < eps
        big = ~small
        return np.concatenate([expm1d_taylor(x[small]),
                               expm1d_naive(x[big])])
    elif abs_x < eps:
        return expm1d_taylor(x)
    else:
        return expm1d_naive(x)

def expm1d_naive(x):
    return np.expm1(x) / x

def expm1d_taylor(x):
    c_n, ret = 1., 1.
    for n in range(2, 11):
        c_n = c_n * x / (1.0 * n)
        ret = ret + c_n
    return ret


def binom_coeffs(n):
    return scipy.special.comb(n, np.arange(n + 1))

log_factorial = lambda n: scipy.special.gammaln(n + 1)
log_binom = lambda n, k: log_factorial(
    n) - log_factorial(k) - log_factorial(n - k)


def hypergeom_mat(N, n):
    K = np.outer(np.ones(n + 1), np.arange(N + 1))
    k = np.outer(np.arange(n + 1), np.ones(N + 1))
    ret = log_binom(K, k)
    ret = ret + ret[::-1, ::-1]
    ret = ret - log_binom(N, n)
    return np.exp(ret)


@memoize
def hypergeom_quasi_inverse(N, n):
    return scipy.linalg.pinv(hypergeom_mat(N, n))


@primitive
def symmetric_matrix(arr, n):
    if len(arr) != n * (n + 1) / 2:
        raise Exception("Array must have dimensions n*(n+1)/2")
    ret = np.zeros((n, n))
    idx = np.triu_indices(n)

    ret[idx] = arr
    ret[tuple(reversed(idx))] = arr

    assert np.all(ret == ret.T)
    return ret
defvjp(symmetric_matrix, lambda ans, arr, n: lambda g: g[np.triu_indices(n)])
#symmetric_matrix.defgrad(lambda ans, arr, n: lambda g: g[np.triu_indices(n)])
#symmetric_matrix.defvjp(lambda g, ans, vs, gvs, arr, n: g[np.triu_indices(n)])


def slogdet_pos(X):
    sgn, slogdet = np.linalg.slogdet(X)
    if sgn <= 0:
        raise Exception("X determinant is nonpositive")
    return slogdet


def log_wishart_pdf(X, V, n, p):
    # correct up to constant of proportionality
    return (n - p - 1) / 2 * slogdet_pos(X) - n / 2 * slogdet_pos(V) - 0.5 * np.trace(np.dot(np.linalg.inv(V), X))


def _apply_error_matrices(vecs, error_matrices):
    if not all([np.allclose(np.sum(err, axis=0), 1.0) for err in error_matrices]):
        raise Exception("Columns of error matrix should sum to 1")

    return [np.dot(v, err) for v, err in zip(vecs, error_matrices)]

# inverse of a PSD matrix


@primitive
def inv_psd(x, **tol_kwargs):
    x = check_psd(x, **tol_kwargs)
    return check_psd(scipy.linalg.pinvh(x), **tol_kwargs)
defvjp(inv_psd, lambda ans, x: lambda g: -np.dot(np.dot(ans, g), ans))
#inv_psd.defgrad(lambda ans, x: lambda g: -np.dot(np.dot(ans, g), ans))
#inv_psd.defvjp(lambda g, ans, vs, gvs, x: -np.dot(np.dot(ans, g), ans))



import numpy as np

import momi.w_matrix

def W(n, b, j):
    if j == 2:
        return 6.0 / (n + 1)
    elif j == 3:
        return 30.0 * (n - 2 * b) / (n + 1) / (n + 2)
    else:
        jj = j - 2
        ret = W(n, b, jj) * -(1 + jj) * (3 + 2 * jj) * \
            (n - jj) / jj / (2 * jj - 1) / (n + jj + 1)
        ret += W(n, b, jj + 1) * (3 + 2 * jj) * (n - 2 * b) / jj / (n + jj + 1)
        return ret


def old_Wmatrix(n):
    ww = np.zeros([n - 1, n - 1])
    for i in range(2, n + 1):
        for j in range(1, n):
            ww[i - 2, j - 1] = W(n, j, i)
    return ww


def test_new_Wmatrix():
    W1 = old_Wmatrix(25)
    W2 = momi.w_matrix.Wmatrix(25)
    np.testing.assert_allclose(W1, W2)



import pytest
import momi
import demo_utils
import io
import collections as co
import subprocess as sp

def test_read_vcf():
    sampled_n_dict = {"a":4,"b":4,"c":6}
    demo = demo_utils.simple_admixture_3pop()
    theta = 100.0
    rho = 100.0
    num_loci = 5
    num_bases = 1e5

    demo.simulate_vcf(
        "test_vcf", recoms_per_gen=rho/num_bases,
        length=num_bases, muts_per_gen=theta/num_bases,
        sampled_n_dict=sampled_n_dict, random_seed=1234,
        force=True)

    data = demo.simulate_data(
        recoms_per_gen=rho/num_bases,
        length=num_bases, muts_per_gen=theta/num_bases,
        sampled_n_dict=sampled_n_dict, num_replicates=1,
        random_seed=1234)

    data2 = momi.SnpAlleleCounts.read_vcf(
        'test_vcf.vcf.gz', ind2pop={f"{pop}_{i}": pop for pop, n in sampled_n_dict.items() for i in range(n)})

    assert data._sfs == data2._sfs.subset_populations(data._sfs.sampled_pops)

    # TODO: test that concatenating datasets from multiple vcfs works?

def test_read_vcf_folded():
    sampled_n_dict = {"a":4,"b":4,"c":6}
    demo = demo_utils.simple_admixture_3pop()
    theta = 100.0
    rho = 100.0
    num_loci = 5
    num_bases = 1e5

    demo.simulate_vcf(
        "test_vcf_folded", recoms_per_gen=rho/num_bases,
        length=num_bases, muts_per_gen=theta/num_bases,
        sampled_n_dict=sampled_n_dict, random_seed=1234,
        force=True, print_aa=False)

    data = demo.simulate_data(
        recoms_per_gen=rho/num_bases,
        length=num_bases, muts_per_gen=theta/num_bases,
        sampled_n_dict=sampled_n_dict, num_replicates=1,
        random_seed=1234)

    data2 = momi.SnpAlleleCounts.read_vcf(
        'test_vcf_folded.vcf.gz', ind2pop={f"{pop}_{i}": pop for pop, n in sampled_n_dict.items() for i in range(n)},
        ancestral_alleles=False)

    assert data._sfs.fold() == data2._sfs.subset_populations(data._sfs.sampled_pops).fold()

    # TODO: test that concatenating datasets from multiple vcfs works?
